# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#from builtins import object
import setup_paths
from nomadcore.simple_parser import SimpleMatcher as SM, mainFunction
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os
import sys
import json


class MopacContext(object):
    """context for the sample parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used
        to parse different files"""
        pass

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext
        # is used to parse different files
        self.initialize_values()

# description of the input

# section_single_configuration_calculation
sm_etot = SM(r"\s*TOTAL ENERGY\s*=\s*(?P<energy_total>[-+0-9.eEdD]+)\s*EV")
sm_eigenvalues = SM(name='Eigenvalues',
                    startReStr=r"\s*EIGENVALUES",
                    endReStr=r"\s*NET ATOMIC CHARGES",
                    sections=['section_eigenvalues'],
                    subMatchers=[
                        SM(r"\s*(?P<eigenvalues_values>[-+0-9.eEdD]+)",
                           repeats=True)
                    ])

sm_energies = SM(name='Energies',
                 startReStr=r"\s*SCF FIELD WAS ACHIEVED",
                 sections=['section_single_configuration_calculation'],
                 subMatchers=[sm_etot, sm_eigenvalues])

# section_method

# section_system

# section_run
#   program_information
sm_version = SM(r"\s*\**\s*Version\s*(?P<program_version>[0-9a-zA-Z_.]*)")
sm_header = SM(name='ProgramHeader',
               startReStr=r"\s*\**\s*Cite this work as:\s*\**",
               subMatchers=[sm_version])

sm_run = SM(name='newRun',
            startReStr=r"",
            #forwardMatch=True,
            required=True,
            fixedStartValues={'program_name': 'mopac'},
            sections=['section_run'],
            subMatchers=[sm_header,
                         sm_energies
                        ])

mainFileDescription = \
    SM(name='root',
       weak=True,
       forwardMatch=True,
       startReStr=r"",
       subMatchers=[sm_run])

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json

parserInfo = {
  "name": "mopac_parser",
  "version": "1.0"
}

class MopacParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('mopac parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("mopac.nomadmetainfo.json")
        context = MopacContext()
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription,
                None,
                parserInfo,
                cachingLevelForMetaName=dict(),
                superContext=context,
                superBackend=backend)

        return backend


if __name__ == "__main__":
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../../../nomad-meta-info/meta_info/nomad_meta_info/mopac.nomadmetainfo.json"))
    metaInfoEnv, warnings = loadJsonFile(filePath=metaInfoPath, dependencyLoader=None, extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS, uri=None)

    superContext = MopacContext()
    mainFunction(mainFileDescription, metaInfoEnv, parserInfo, superContext=superContext)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The encyclopedia API of the nomad@FAIRDI APIs.
"""
import re
import math
import numpy as np

from flask_restplus import Resource, abort, fields, marshal
from flask import request, g
from elasticsearch_dsl import Search, Q, A
from elasticsearch_dsl.utils import AttrDict

from nomad import config, infrastructure, search
from nomad.files import UploadFiles
from nomad.units import ureg
from nomad.atomutils import get_hill_decomposition
from nomad.datamodel.datamodel import EntryArchive
from nomad.datamodel.material import Material
from .api import api
from .auth import authenticate, create_authorization_predicate

ns = api.namespace("encyclopedia", description="Access encyclopedia metadata.")
re_formula = re.compile(r"([A-Z][a-z]?)(\d*)")

material_prop_map = {
    # General
    "material_id": "encyclopedia.material.material_id",
    "formula": "encyclopedia.material.formula",
    "formula_reduced": "encyclopedia.material.formula_reduced",
    "system_type": "encyclopedia.material.material_type",
    # Bulk only
    "has_free_wyckoff_parameters": "encyclopedia.material.bulk.has_free_wyckoff_parameters",
    "strukturbericht_designation": "encyclopedia.material.bulk.strukturbericht_designation",
    "material_name": "encyclopedia.material.material_name",
    "bravais_lattice": "encyclopedia.material.bulk.bravais_lattice",
    "crystal_system": "encyclopedia.material.bulk.crystal_system",
    "point_group": "encyclopedia.material.bulk.point_group",
    "space_group_number": "encyclopedia.material.bulk.space_group_number",
    "space_group_international_short_symbol": "encyclopedia.material.bulk.space_group_international_short_symbol",
    "structure_prototype": "encyclopedia.material.bulk.structure_prototype",
    "structure_type": "encyclopedia.material.bulk.structure_type",
}


def rgetattr(obj, attr_name):
    """Used to perform attribute access based on a (possibly nested) attribute
    name given as string.
    """
    try:
        for attr in attr_name.split("."):
            obj = obj[attr]
    except KeyError:
        return None
    return obj


def get_es_doc_values(es_doc, mapping, keys=None):
    """Used to form a material definition for "materials/<material_id>" from
    the given ElasticSearch root document.
    """
    if keys is None:
        keys = mapping.keys()

    result = {}
    for key in keys:
        es_key = mapping[key]
        value = rgetattr(es_doc, es_key)
        result[key] = value

    return result


def get_enc_filter():
    """Returns a shared term filter that will leave out unpublished (of other
    users), embargoed or invalid entries.
    """
    # Handle authentication
    s = search.SearchRequest()
    if g.user is not None:
        s.owner('visible', user_id=g.user.user_id)
    else:
        s.owner('public')
    return [
        s.q,
        Q("term", encyclopedia__status="success"),
    ]


similarity = api.model("similarity", {
    # General
    "material_id": fields.String,
    "value": fields.Float,
    "formula": fields.String,
    "space_group_number": fields.Integer,
})
material_query = api.parser()
material_query.add_argument(
    "property",
    type=str,
    choices=tuple(material_prop_map.keys()),
    help="Optional single property to retrieve for the given material. If not specified, all properties will be returned.",
    location="args"
)
material_result = api.model("material_result", {
    # General
    "material_id": fields.String,
    "formula": fields.String,
    "formula_reduced": fields.String,
    "system_type": fields.String,
    "n_matches": fields.Integer,
    # Bulk only
    "has_free_wyckoff_parameters": fields.Boolean,
    "strukturbericht_designation": fields.String,
    "material_name": fields.String,
    "bravais_lattice": fields.String,
    "crystal_system": fields.String,
    "point_group": fields.String,
    "space_group_number": fields.Integer,
    "space_group_international_short_symbol": fields.String,
    "structure_prototype": fields.String,
    "structure_type": fields.String,
    "similarity": fields.List(fields.Nested(similarity, skip_none=True), skip_none=True),
})


@ns.route("/materials/<string:material_id>")
class EncMaterialResource(Resource):
    @api.response(404, "The material does not exist")
    @api.response(200, "Metadata send", fields.Raw)
    @api.doc("get_material")
    @api.expect(material_query)
    @api.marshal_with(material_result, skip_none=True)
    @authenticate()
    def get(self, material_id):
        """Used to retrieve basic information related to the specified
        material.
        """
        # Parse request arguments
        args = material_query.parse_args()
        prop = args.get("property", None)
        if prop is not None:
            keys = [prop]
            es_keys = [material_prop_map[prop]]
        else:
            keys = list(material_prop_map.keys())
            es_keys = list(material_prop_map.values())

        # Find the first public entry with this material id and take
        # information from there. In principle all other entries should have
        # the same information.
        s = Search(index=config.elastic.index_name)
        query = Q(
            "bool",
            filter=get_enc_filter() + [
                Q("term", encyclopedia__material__material_id=material_id),
            ]
        )
        s = s.query(query)

        # Only one representative entry is returned by collapsing the results.
        s = s.extra(**{
            "_source": {"includes": es_keys},
            "size": 1,
            "collapse": {"field": "encyclopedia.material.material_id"},
        })
        response = s.execute()

        # No such material
        if len(response) == 0:
            abort(404, message="There is no material {}".format(material_id))

        # Add values from ES entry
        entry = response[0]
        result = get_es_doc_values(entry, material_prop_map, keys)

        # Add similarity data that is currently stored in MongoDB. In the
        # future a lot of the data will be accessed here.
        try:
            material = Material.m_def.a_mongo.get(material_id=material_id)
            dos_similarity = material.similarity.electronic_dos
        except KeyError:
            # No similarity data for this material
            pass
        else:
            # Only include similarity for materials that exist on the current
            # deployment to avoid dead links.
            similar_ids = dos_similarity.material_ids
            id_value_map = {key: value for key, value in zip(dos_similarity.material_ids, dos_similarity.values)}
            bool_query = Q(
                "bool",
                filter=get_enc_filter() + [Q("terms", encyclopedia__material__material_id=similar_ids)],
            )
            s = Search(index=config.elastic.index_name)
            s = s.query(bool_query)
            s = s.extra(**{
                "_source": {"includes": [
                    "encyclopedia.material.material_id",
                    "encyclopedia.material.formula_reduced",
                    "encyclopedia.material.bulk.space_group_number",
                ]},
                "size": 5,
                "collapse": {"field": "encyclopedia.material.material_id"},
            })
            response = s.execute()
            similarity = []
            for hit in response.hits:
                try:
                    similarity.append({
                        "material_id": hit.encyclopedia.material.material_id,
                        "value": id_value_map[hit.encyclopedia.material.material_id],
                        "formula": hit.encyclopedia.material.formula_reduced,
                        "space_group_number": hit.encyclopedia.material.bulk.space_group_number,
                    })
                except AttributeError:
                    pass
            if similarity:
                result["similarity"] = similarity

        return result, 200


range_query = api.model("range_query", {
    "max": fields.Float,
    "min": fields.Float,
})
materials_after = api.model("materials_after", {
    "materials": fields.String,
})
materials_query = api.model("materials_input", {
    "search_by": fields.Nested(api.model("search_query", {
        "exclusive": fields.Boolean(default=False),
        "formula": fields.String,
        "element": fields.String,
        "page": fields.Integer(default=1),
        "after": fields.Nested(materials_after, allow_null=True),
        "per_page": fields.Integer(default=25),
        "pagination": fields.Boolean,
    })),
    "material_name": fields.List(fields.String),
    "structure_type": fields.List(fields.String),
    "space_group_number": fields.List(fields.Integer),
    "system_type": fields.List(fields.String),
    "crystal_system": fields.List(fields.String),
    "band_gap": fields.Nested(range_query, description="Band gap range in eV."),
    "band_gap_direct": fields.Boolean,
    "has_band_structure": fields.Boolean,
    "has_dos": fields.Boolean,
    "has_fermi_surface": fields.Boolean,
    "has_thermal_properties": fields.Boolean,
    "functional_type": fields.List(fields.String),
    "basis_set_type": fields.List(fields.String),
    "code_name": fields.List(fields.String),
    "mass_density": fields.Nested(range_query, description="Mass density range in kg / m ** 3."),
})
pages_result = api.model("page_info", {
    "per_page": fields.Integer,
    "total": fields.Integer,
    "page": fields.Integer,
    "pages": fields.Integer,
    "after": fields.Nested(materials_after),
})

materials_result = api.model("materials_result", {
    "total_results": fields.Integer(allow_null=False),
    "results": fields.List(fields.Nested(material_result, skip_none=True)),
    "pages": fields.Nested(pages_result, skip_none=True),
    "es_query": fields.String(allow_null=False),
})


@ns.route("/materials")
class EncMaterialsResource(Resource):
    @api.response(404, "No materials found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.expect(materials_query, validate=False)
    @api.marshal_with(materials_result, skip_none=True)
    @api.doc("materials")
    @authenticate()
    def post(self):
        """Used to query a list of materials with the given search options.
        """
        # Get query parameters as json
        try:
            data = marshal(request.get_json(), materials_query)
        except Exception as e:
            abort(400, message=str(e))

        def add_terms_filter(filters, source, target, query_type="terms"):
            """For adding terms filters
            """
            if data[source] is not None:
                filters.append(Q(query_type, **{target: data[source]}))

        def add_exists_filter(filters, must_nots, source, target):
            """For adding exists filters
            """
            param = data[source]
            if param is not None:
                query = Q("exists", field=target)
                if param is True:
                    filters.append(query)
                elif param is False:
                    must_nots.append(query)

        def add_range_filter(filters, source, target, source_unit=None, target_unit=None):
            """For adding range filters
            """
            param = data[source]
            query_dict = {}
            if param["min"] is not None:
                if source_unit is None and target_unit is None:
                    gte = param["min"]
                else:
                    gte = (param["min"] * source_unit).to(target_unit).magnitude
                query_dict["gte"] = gte
            if param["max"] is not None:
                if source_unit is None and target_unit is None:
                    lte = param["max"]
                else:
                    lte = (param["max"] * source_unit).to(target_unit).magnitude
                query_dict["lte"] = lte
            if len(query_dict) != 0:
                query = Q("range", **{target: query_dict})
                filters.append(query)

        property_map = {
            "has_thermal_properties": "encyclopedia.properties.thermodynamical_properties",
            "has_band_structure": "encyclopedia.properties.electronic_band_structure",
            "has_dos": "encyclopedia.properties.electronic_dos",
            "has_fermi_surface": "encyclopedia.properties.fermi_surface",
        }
        requested_properties = []
        filters = get_enc_filter()
        must_nots = []
        musts = []
        add_terms_filter(filters, "material_name", "encyclopedia.material.material_name")
        add_terms_filter(filters, "structure_type", "encyclopedia.material.bulk.structure_type")
        add_terms_filter(filters, "space_group_number", "encyclopedia.material.bulk.space_group_number")
        add_terms_filter(filters, "system_type", "encyclopedia.material.material_type")
        add_terms_filter(filters, "crystal_system", "encyclopedia.material.bulk.crystal_system")
        add_terms_filter(filters, "band_gap_direct", "encyclopedia.properties.band_gap_direct", query_type="term")
        add_terms_filter(filters, "functional_type", "encyclopedia.method.functional_type")
        add_terms_filter(filters, "basis_set_type", "dft.basis_set")
        add_terms_filter(filters, "code_name", "dft.code_name")
        add_range_filter(filters, "band_gap", "encyclopedia.properties.band_gap", ureg.eV, ureg.J)
        add_range_filter(filters, "mass_density", "encyclopedia.properties.mass_density")

        # Create query for elements or formula
        search_by = data["search_by"]
        formula = search_by["formula"]
        elements = search_by["element"]
        exclusive = search_by["exclusive"]

        if formula is not None:
            # Here we determine a list of atom types. The types may occur
            # multiple times and at multiple places.
            element_list = []
            matches = re_formula.finditer(formula)
            for match in matches:
                groups = match.groups()
                symbol = groups[0]
                count = groups[1]
                if symbol != "":
                    if count == "":
                        element_list.append(symbol)
                    else:
                        element_list += [symbol] * int(count)

            # The given list of species is reformatted with the Hill system
            # into a query string. The counts are reduced by the greatest
            # common divisor.
            names, reduced_counts = get_hill_decomposition(element_list, reduced=True)
            query_string = []
            for name, count in zip(names, reduced_counts):
                if count == 1:
                    query_string.append(name)
                else:
                    query_string.append("{}{}".format(name, int(count)))
            query_string = " ".join(query_string)

            # With exclusive search we look for exact match
            if exclusive:
                filters.append(Q("term", **{"encyclopedia.material.species_and_counts.keyword": query_string}))
            # With non-exclusive search we look for match that includes at
            # least all parts of the formula, possibly even more.
            else:
                musts.append(Q(
                    "match",
                    encyclopedia__material__species_and_counts={"query": query_string, "operator": "and"}
                ))
        elif elements is not None:
            # The given list of species is reformatted with the Hill system into a query string
            species, _ = get_hill_decomposition(elements.split(","))
            query_string = " ".join(species)

            # With exclusive search we look for exact match
            if exclusive:
                filters.append(Q("term", **{"encyclopedia.material.species.keyword": query_string}))
            # With non-exclusive search we look for match that includes at
            # least all species, possibly even more.
            else:
                musts.append(Q(
                    "match",
                    encyclopedia__material__species={"query": query_string, "operator": "and"}
                ))

        # The queries that correspond to AND queries typically need to access
        # multiple calculations at once to find the material ids that
        # correspond to the query. To implement this behaviour we need to run
        # an initial aggregation that checks that the requested properties are
        # present for a material. This is a a very crude solution that does not
        # scale to complex queries, but I'm not sure we can do much better
        # until we have a separate index for materials. The size is set very
        # large because all the results need to be returned. We cannot get the
        # results in a paginated way with composite aggregation, because
        # pipeline aggregations are not compatible with them.
        agg_parent = A("terms", field="encyclopedia.material.material_id", size=500000)
        for key, value in property_map.items():
            if data[key] is True:
                agg = A("filter", exists={"field": value})
                agg_parent.bucket(key, agg)
                requested_properties.append(key)
        if len(requested_properties) > 1:
            # First we setup a boolean filter query that filters for of the
            # requested properties. This will reduce the size of the initial
            # set on top of which the more expensive aggregation stack is run
            # on.
            bool_query = Q(
                "bool",
                filter=filters,
                must_not=must_nots,
                must=musts,
                should=[Q("exists", field=property_map[x]) for x in requested_properties],
                minimum_should_match=1,  # At least one of the should query must match
            )
            s = Search(index=config.elastic.index_name)
            s = s.query(bool_query)

            # The remaining requested properties have to be queried as a nested
            # aggregation.
            s.aggs.bucket("materials", agg_parent)
            buckets_path = {x: "{}._count".format(x) for x in requested_properties}
            script = " && ".join(["params.{} > 0".format(x) for x in requested_properties])
            agg_parent.pipeline("selector", A(
                "bucket_selector",
                buckets_path=buckets_path,
                script=script,
            ))
            s = s.extra(**{
                "size": 0,
            })
            response = s.execute()
            material_ids = [x["key"] for x in response.aggs.materials.buckets]
            if len(material_ids) == 0:
                abort(404, message="No materials found for the given search criteria or pagination.")

        # Add pre-selected material ids if multiple exists filters were
        # requested. These IDs are already filtered based on the user query so
        # none of the other search terms need be used.
        if len(requested_properties) > 1:
            must_nots = []
            musts = []
            filters = []
            filters.append(Q("terms", encyclopedia__material__material_id=material_ids))
        if len(requested_properties) == 1:
            prop_name = requested_properties[0]
            add_exists_filter(filters, must_nots, prop_name, property_map[prop_name])

        # The top query filters out entries based on the user query
        page = search_by["page"]
        per_page = search_by["per_page"]
        after = search_by["after"]
        bool_query = Q(
            "bool",
            filter=filters,
            must_not=must_nots,
            must=musts,
        )
        s = Search(index=config.elastic.index_name)
        s = s.query(bool_query)

        # 1: The paginated approach: No way to know the amount of materials,
        # but can return aggregation results in a quick fashion including
        # the number of calculation entries per material.
        mode = "collapse"
        if mode == "aggregation":
            # The materials are grouped by using three aggregations:
            # "Composite" to enable scrolling, "Terms" to enable selecting
            # by material_id and "Top Hits" to fetch a single
            # representative material document. Unnecessary fields are
            # filtered to reduce data transfer.
            terms_agg = A("terms", field="encyclopedia.material.material_id")
            composite_kwargs = {"sources": {"materials": terms_agg}, "size": per_page}

            # The number of matched materials is only requested on the first
            # search, not for each page.
            if after is not None:
                composite_kwargs["after"] = after
            else:
                cardinality_agg = A("cardinality", field="encyclopedia.material.material_id", precision_threshold=1000)
                s.aggs.metric("n_materials", cardinality_agg)

            composite_agg = A("composite", **composite_kwargs)
            composite_agg.metric("representative", A(
                "top_hits",
                size=1,
                _source={"includes": list(material_prop_map.values())},
            ))
            s.aggs.bucket("materials", composite_agg)

            # We ignore the top level hits and sort by reduced material formula.
            s = s.extra(**{
                "size": 0,
            })

            response = s.execute()
            materials = response.aggs.materials.buckets
            if len(materials) == 0:
                abort(404, message="No materials found for the given search criteria or pagination.")
            after_new = response.aggs.materials["after_key"]

            # Gather results from aggregations
            result_list = []
            materials = response.aggs.materials.buckets
            keys = list(material_prop_map.keys())
            for material in materials:
                representative = material["representative"][0]
                mat_dict = get_es_doc_values(representative, material_prop_map, keys)
                mat_dict["n_matches"] = material.doc_count
                result_list.append(mat_dict)

            # Page information is incomplete for aggregations
            pages = {
                "page": page,
                "per_page": per_page,
                "after": after_new,
            }
            if after is None:
                n_materials = response.aggs.n_materials.value
                pages["total"] = n_materials

        # 2. Collapse approach. Quickly provides a list of materials
        # corresponding to the query, offers full pagination, the number of
        # matches per material needs to be requested with a separate query.
        elif mode == "collapse":
            s = Search(index=config.elastic.index_name)
            s = s.query(bool_query)

            # Add cardinality aggregation that gives out the total number of materials
            cardinality_agg = A("cardinality", field="encyclopedia.material.material_id", precision_threshold=1000)
            s.aggs.metric("n_materials", cardinality_agg)

            s = s.extra(**{
                "collapse": {"field": "encyclopedia.material.material_id"},
                "size": per_page,
                "from": (page - 1) * per_page,
                "sort": [{"encyclopedia.material.formula_reduced": {"order": "asc"}}],
                "explain": True,
            })

            # Execute query
            response = s.execute()

            # No matches
            if len(response) == 0:
                abort(404, message="No materials found for the given search criteria or pagination.")

            # Gather number of entries per material with a separate query
            material_ids = [x.encyclopedia.material.material_id for x in response]
            s = Search(index=config.elastic.index_name)
            bool_query = Q(
                "bool",
                filter=Q("terms", encyclopedia__material__material_id=material_ids),
            )
            s2 = s.query(bool_query)
            s2.aggs.bucket("n_matches", A("terms", field="encyclopedia.material.material_id"))
            response2 = s2.execute()
            matmap = {x.key: x.doc_count for x in response2.aggs.n_matches}

            # Loop over materials
            result_list = []
            keys = list(material_prop_map.keys())
            for material in response:
                # Get values from the collapsed doc
                mat_result = get_es_doc_values(material, material_prop_map, keys)
                mat_id = material.encyclopedia.material.material_id
                mat_result["n_matches"] = matmap[mat_id]
                result_list.append(mat_result)

            # Full page information available for collapse
            pages = {
                "page": page,
                "per_page": per_page,
                "pages": math.ceil(response.hits.total / per_page),
                "total": response.aggs.n_materials.value,
            }

        result = {
            "results": result_list,
            "pages": pages,
        }
        return result, 200


groups_result = api.model("groups_result", {
    "groups_eos": fields.Raw,
    "groups_par": fields.Raw,
})


@ns.route("/materials/<string:material_id>/groups")
class EncGroupsResource(Resource):
    @api.response(404, "Material not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.marshal_with(groups_result)
    @api.doc("enc_materials")
    @authenticate()
    def get(self, material_id):
        """Returns a summary of the calculation groups that were identified for
        this material.
        """
        # Find entries for the given material, which have EOS or parameter
        # variation hashes set.
        bool_query = Q(
            "bool",
            filter=get_enc_filter() + [Q("term", encyclopedia__material__material_id=material_id)],
            must=[
                Q("exists", field="encyclopedia.properties.energies.energy_total"),
                Q("exists", field="encyclopedia.material.idealized_structure.cell_volume"),
            ],
            should=[
                Q("exists", field="encyclopedia.method.group_eos_id"),
                Q("exists", field="encyclopedia.method.group_parametervariation_id"),
            ],
            minimum_should_match=1,  # At least one of the should query must match
        )

        s = Search(index=config.elastic.index_name)
        s = s.query(bool_query)

        # Bucket the calculations by the group hashes. Only create a bucket if an
        # above-minimum number of documents are found.
        group_eos_bucket = A("terms", field="encyclopedia.method.group_eos_id", min_doc_count=4)
        group_param_bucket = A("terms", field="encyclopedia.method.group_parametervariation_id", min_doc_count=2)
        calc_aggregation = A(
            "top_hits",
            _source={"includes": ["calc_id"]},
            sort=[{"encyclopedia.properties.energies.energy_total": {"order": "asc"}}],
            size=100,
        )
        group_eos_bucket.bucket("calculations", calc_aggregation)
        group_param_bucket.bucket("calculations", calc_aggregation)
        s.aggs.bucket("groups_eos", group_eos_bucket)
        s.aggs.bucket("groups_param", group_param_bucket)

        # We ignore the top level hits
        s = s.extra(**{
            "size": 0,
        })

        # Collect information for each group from the aggregations
        response = s.execute()
        groups_eos = {group.key: [calc.calc_id for calc in group.calculations.hits] for group in response.aggs.groups_eos.buckets}
        groups_param = {group.key: [calc.calc_id for calc in group.calculations.hits] for group in response.aggs.groups_param.buckets}

        # Return results
        result = {
            "groups_eos": groups_eos,
            "groups_par": groups_param,
        }

        return result, 200


group_result = api.model("group_result", {
    "calculations": fields.List(fields.String),
    "energies": fields.List(fields.Float),
    "volumes": fields.List(fields.Float),
})
group_source = {
    "includes": [
        "calc_id",
        "encyclopedia.properties.energies.energy_total",
        "encyclopedia.material.idealized_structure.cell_volume",
    ]
}


@ns.route("/materials/<string:material_id>/groups/<string:group_type>/<string:group_id>")
class EncGroupResource(Resource):
    @api.response(404, "Group not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.marshal_with(group_result)
    @api.doc("enc_group")
    @authenticate()
    def get(self, material_id, group_type, group_id):
        """Used to query detailed information for a specific calculation group.
        """
        # Find entries for the given material, which have EOS or parameter
        # variation hashes set.
        if group_type == "eos":
            group_id_source = "encyclopedia.method.group_eos_id"
        elif group_type == "par":
            group_id_source = "encyclopedia.method.group_parametervariation_id"
        else:
            abort(400, message="Unsupported group type.")

        bool_query = Q(
            "bool",
            filter=get_enc_filter() + [
                Q("term", encyclopedia__material__material_id=material_id),
                Q("term", **{group_id_source: group_id}),
            ],
        )

        s = Search(index=config.elastic.index_name)
        s = s.query(bool_query)

        # calc_id and energy should be extracted for each matched document. The
        # documents are sorted by energy so that the minimum energy one can be
        # easily extracted. A maximum request size is set in order to limit the
        # result size. ES also has an index-level property
        # "index.max_inner_result_window" that limits the number of results
        # that an inner result can contain.
        energy_aggregation = A(
            "top_hits",
            _source=group_source,
            sort=[{"encyclopedia.properties.energies.energy_total": {"order": "asc"}}],
            size=100,
        )
        s.aggs.bucket("groups_eos", energy_aggregation)

        # We ignore the top level hits
        s = s.extra(**{
            "size": 0,
        })

        # Collect information for each group from the aggregations
        response = s.execute()

        hits = response.aggs.groups_eos.hits
        calculations = [doc.calc_id for doc in hits]
        energies = [doc.encyclopedia.properties.energies.energy_total for doc in hits]
        volumes = [doc.encyclopedia.material.idealized_structure.cell_volume for doc in hits]
        group_dict = {
            "calculations": calculations,
            "energies": energies,
            "volumes": volumes,
        }

        return group_dict, 200


suggestions_map = {
    "code_name": "dft.code_name",
    "structure_type": "encyclopedia.material.bulk.structure_type",
}
suggestions_query = api.parser()
suggestions_query.add_argument(
    "property",
    type=str,
    choices=("code_name", "structure_type"),
    help="The property name for which suggestions are returned.",
    location="args"
)
suggestions_result = api.model("suggestions_result", {
    "code_name": fields.List(fields.String),
    "structure_type": fields.List(fields.String),
})


@ns.route("/suggestions")
class EncSuggestionsResource(Resource):
    @api.response(404, "Suggestion not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.expect(suggestions_query, validate=False)
    @api.marshal_with(suggestions_result, skip_none=True)
    @api.doc("enc_suggestions")
    @authenticate()
    def get(self):

        # Parse request arguments
        args = suggestions_query.parse_args()
        prop = args.get("property", None)

        # Use aggregation to return all unique terms for the requested field.
        # Without using composite aggregations there is a size limit for the
        # number of aggregation buckets. This should, however, not be a problem
        # since the number of unique values is low for all supported properties.
        s = Search(index=config.elastic.index_name)
        query = Q(
            "bool",
            filter=get_enc_filter()
        )
        s = s.query(query)
        s = s.extra(**{
            "size": 0,
        })

        terms_agg = A("terms", field=suggestions_map[prop])
        s.aggs.bucket("suggestions", terms_agg)

        # Gather unique values into a list
        response = s.execute()
        suggestions = [x.key for x in response.aggs.suggestions.buckets]

        return {prop: suggestions}, 200


calc_prop_map = {
    "calc_id": "calc_id",
    "upload_id": "upload_id",
    "code_name": "dft.code_name",
    "code_version": "dft.code_version",
    "functional_type": "encyclopedia.method.functional_type",
    "basis_set_type": "dft.basis_set",
    "core_electron_treatment": "encyclopedia.method.core_electron_treatment",
    "run_type": "encyclopedia.calculation.calculation_type",
    "has_dos": "encyclopedia.properties.electronic_dos",
    "has_band_structure": "encyclopedia.properties.electronic_band_structure",
    "has_thermal_properties": "encyclopedia.properties.thermodynamical_properties",
    "has_phonon_dos": "encyclopedia.properties.phonon_dos",
    "has_phonon_band_structure": "encyclopedia.properties.phonon_band_structure",
}
calculation_result = api.model("calculation_result", {
    "calc_id": fields.String,
    "upload_id": fields.String,
    "code_name": fields.String,
    "code_version": fields.String,
    "functional_type": fields.String,
    "basis_set_type": fields.String,
    "core_electron_treatment": fields.String,
    "run_type": fields.String,
    "has_dos": fields.Boolean,
    "has_band_structure": fields.Boolean,
    "has_thermal_properties": fields.Boolean,
    "has_phonon_dos": fields.Boolean,
    "has_phonon_band_structure": fields.Boolean,
})
representatives_result = api.model("representatives_result", {
    "idealized_structure": fields.String,
    "electronic_band_structure": fields.String,
    "electronic_dos": fields.String,
    "thermodynamical_properties": fields.String,
})
calculations_result = api.model("calculations_result", {
    "total_results": fields.Integer,
    "pages": fields.Nested(pages_result),
    "results": fields.List(fields.Nested(calculation_result)),
    "representatives": fields.Nested(representatives_result, skip_none=True),
})


@ns.route("/materials/<string:material_id>/calculations")
class EncCalculationsResource(Resource):
    @api.response(404, "Suggestion not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.doc("get_calculations")
    @authenticate()
    def get(self, material_id):
        """Used to return all calculations related to the given material. Also
        returns a representative calculation for each property shown in the
        overview page.
        """
        s = Search(index=config.elastic.index_name)
        query = Q(
            "bool",
            filter=get_enc_filter() + [
                Q("term", encyclopedia__material__material_id=material_id),
            ]
        )
        s = s.query(query)

        # The query is filtered already on the ES side so we don"t need to
        # transfer so much data.
        s = s.extra(**{
            "_source": {"includes": list(calc_prop_map.values()) + ["dft.xc_functional"]},
            "size": 10000,
            "from": 0,
        })
        response = s.execute()

        # No such material
        if len(response) == 0:
            abort(404, message="There is no material {}".format(material_id))

        # Add representative properties. It might be possible to write a custom
        # ES scoring mechanism or aggregation to also perform the selection.
        representatives = {}

        def calc_score(entry):
            """Custom scoring function used to sort results by their
            "quality". Currently built to mimic the scoring that was used
            in the old Encyclopedia GUI. Primarily sorts by quality measure,
            ties are broken by alphabetic sorting of entry_id in order to
            return consistent results.
            """
            score = 0
            functional_score = {
                "GGA": 100
            }
            code_score = {
                "FHI-aims": 3,
                "VASP": 2,
                "Quantum Espresso": 1,
            }
            code_name = entry.dft.code_name
            functional = entry.dft.xc_functional
            has_dos = rgetattr(entry, "encyclopedia.properties.electronic_band_structure") is not None
            has_bs = rgetattr(entry, "encyclopedia.properties.electronic_dos") is not None
            score += functional_score.get(functional, 0)
            score += code_score.get(code_name, 0)
            if has_dos and has_bs:
                score += 10

            return (score, entry["calc_id"])

        # The calculations are first sorted by "quality"
        sorted_calc = sorted(response, key=lambda x: calc_score(x), reverse=True)

        # Get the requested representative properties
        representatives["idealized_structure"] = sorted_calc[0].calc_id
        thermo_found = False
        bs_found = False
        dos_found = False
        for calc in sorted_calc:
            if rgetattr(calc, "encyclopedia.properties.thermodynamical_properties") is not None:
                representatives["thermodynamical_properties"] = calc.calc_id
                thermo_found = True
            if rgetattr(calc, "encyclopedia.properties.electronic_band_structure") is not None:
                representatives["electronic_band_structure"] = calc.calc_id
                bs_found = True
            if rgetattr(calc, "encyclopedia.properties.electronic_dos") is not None:
                representatives["electronic_dos"] = calc.calc_id
                dos_found = True
            if thermo_found and bs_found and dos_found:
                break

        # Create result JSON
        results = []
        for entry in response:
            calc_dict = get_es_doc_values(entry, calc_prop_map)
            calc_dict["has_dos"] = calc_dict["has_dos"] is not None
            calc_dict["has_band_structure"] = calc_dict["has_band_structure"] is not None
            calc_dict["has_thermal_properties"] = calc_dict["has_thermal_properties"] is not None
            calc_dict["has_phonon_dos"] = calc_dict["has_phonon_dos"] is not None
            calc_dict["has_phonon_band_structure"] = calc_dict["has_phonon_band_structure"] is not None
            results.append(calc_dict)

        result = {
            "total_results": len(results),
            "results": results,
            "representatives": representatives,
        }

        return result, 200


histogram = api.model("histogram", {
    "occurrences": fields.List(fields.Integer),
    "values": fields.List(fields.Float),
})
statistics_query = api.model("statistics_query", {
    "calculations": fields.List(fields.String),
    "properties": fields.List(fields.String),
    "n_histogram_bins": fields.Integer,
})
statistics = api.model("statistics", {
    "min": fields.Float,
    "max": fields.Float,
    "avg": fields.Float,
    "histogram": fields.Nested(histogram, skip_none=True)
})
statistics_result = api.model("statistics_result", {
    "cell_volume": fields.Nested(statistics, skip_none=True),
    "atomic_density": fields.Nested(statistics, skip_none=True),
    "mass_density": fields.Nested(statistics, skip_none=True),
    "lattice_a": fields.Nested(statistics, skip_none=True),
    "lattice_b": fields.Nested(statistics, skip_none=True),
    "lattice_c": fields.Nested(statistics, skip_none=True),
    "alpha": fields.Nested(statistics, skip_none=True),
    "beta": fields.Nested(statistics, skip_none=True),
    "gamma": fields.Nested(statistics, skip_none=True),
    "band_gap": fields.Nested(statistics, skip_none=True),
})
property_map = {
    "cell_volume": "encyclopedia.material.idealized_structure.cell_volume",
    "atomic_density": "encyclopedia.properties.atomic_density",
    "mass_density": "encyclopedia.properties.mass_density",
    "lattice_a": "encyclopedia.material.idealized_structure.lattice_parameters.a",
    "lattice_b": "encyclopedia.material.idealized_structure.lattice_parameters.b",
    "lattice_c": "encyclopedia.material.idealized_structure.lattice_parameters.c",
    "alpha": "encyclopedia.material.idealized_structure.lattice_parameters.alpha",
    "beta": "encyclopedia.material.idealized_structure.lattice_parameters.beta",
    "gamma": "encyclopedia.material.idealized_structure.lattice_parameters.gamma",
    "band_gap": "encyclopedia.properties.band_gap",
}


@ns.route("/materials/<string:material_id>/statistics")
class EncStatisticsResource(Resource):
    @api.response(404, "Suggestion not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.expect(statistics_query, validate=False)
    @api.marshal_with(statistics_result, skip_none=True)
    @api.doc("enc_statistics")
    @authenticate()
    def post(self, material_id):
        """Used to return statistics related to the specified material and
        calculations.
        """
        # Get query parameters as json
        try:
            data = marshal(request.get_json(), statistics_query)
        except Exception as e:
            abort(400, message=str(e))

        # Find entries for the given material.
        bool_query = Q(
            "bool",
            filter=get_enc_filter() + [
                Q("term", encyclopedia__material__material_id=material_id),
                Q("terms", calc_id=data["calculations"]),
            ]
        )

        s = Search(index=config.elastic.index_name)
        s = s.query(bool_query)
        s = s.extra(**{
            "size": 0,
        })

        # Add statistics aggregations for each requested property
        properties = data["properties"]
        for prop in properties:
            stats_agg = A("stats", field=property_map[prop])
            s.aggs.bucket("{}_stats".format(prop), stats_agg)

        # No hits on the top query level
        response = s.execute()
        if response.hits.total == 0:
            abort(404, message="Could not find matching calculations.")

        # Run a second query that creates histograms with fixed size buckets
        # based on the min and max from previous query. Might make more sense
        # to use the mean and sigma to define the range?
        s = Search(index=config.elastic.index_name)
        s = s.query(bool_query)
        s = s.extra(**{
            "size": 0,
        })
        n_bins = data["n_histogram_bins"]
        for prop in properties:
            stats = getattr(response.aggs, "{}_stats".format(prop))
            if stats.count == 0:
                continue
            interval = (stats.max * 1.001 - stats.min) / n_bins
            if interval == 0:
                interval = 1
            hist_agg = A("histogram", field=property_map[prop], interval=interval, offset=stats.min, min_doc_count=0)
            s.aggs.bucket("{}_hist".format(prop), hist_agg)
        response_hist = s.execute()

        # Return results
        result = {}
        for prop in properties:
            stats = getattr(response.aggs, "{}_stats".format(prop))
            if stats.count == 0:
                continue
            hist = getattr(response_hist.aggs, "{}_hist".format(prop))
            occurrences = [x.doc_count for x in hist.buckets]
            values = [x.key for x in hist.buckets]
            result[prop] = {
                "min": stats.min,
                "max": stats.max,
                "avg": stats.avg,
                "histogram": {
                    "occurrences": occurrences,
                    "values": values,
                }
            }

        return result, 200


wyckoff_variables_result = api.model("wyckoff_variables_result", {
    "x": fields.Float,
    "y": fields.Float,
    "z": fields.Float,
})
wyckoff_set_result = api.model("wyckoff_set_result", {
    "wyckoff_letter": fields.String,
    "indices": fields.List(fields.Integer),
    "element": fields.String,
    "variables": fields.Nested(wyckoff_variables_result, skip_none=True),
})
lattice_parameters = api.model("lattice_parameters", {
    "a": fields.Float,
    "b": fields.Float,
    "c": fields.Float,
    "alpha": fields.Float,
    "beta": fields.Float,
    "gamma": fields.Float,
})

idealized_structure_result = api.model("idealized_structure_result", {
    "atom_labels": fields.List(fields.String),
    "atom_positions": fields.List(fields.List(fields.Float)),
    "lattice_vectors": fields.List(fields.List(fields.Float)),
    "lattice_vectors_primitive": fields.List(fields.List(fields.Float)),
    "lattice_parameters": fields.Nested(lattice_parameters, skip_none=True),
    "periodicity": fields.List(fields.Boolean),
    "number_of_atoms": fields.Integer,
    "cell_volume": fields.Float,
    "wyckoff_sets": fields.List(fields.Nested(wyckoff_set_result, skip_none=True)),
})

calculation_property_map = {
    "lattice_parameters": {
        "es_source": "encyclopedia.material.idealized_structure.lattice_parameters"
    },
    "energies": {
        "es_source": "encyclopedia.properties.energies",
    },
    "mass_density": {
        "es_source": "encyclopedia.properties.mass_density",
    },
    "atomic_density": {
        "es_source": "encyclopedia.properties.atomic_density",
    },
    "cell_volume": {
        "es_source": "encyclopedia.material.idealized_structure.cell_volume"
    },
    "band_gap": {
        "es_source": "encyclopedia.properties.band_gap"
    },
    "electronic_band_structure": {
        "es_source": "encyclopedia.properties.electronic_band_structure"
    },
    "electronic_dos": {
        "es_source": "encyclopedia.properties.electronic_dos"
    },
    "phonon_band_structure": {
        "es_source": "encyclopedia.properties.phonon_band_structure"
    },
    "phonon_dos": {
        "es_source": "encyclopedia.properties.phonon_dos"
    },
    "thermodynamical_properties": {
        "es_source": "encyclopedia.properties.thermodynamical_properties"
    },
    "wyckoff_sets": {
        "arch_source": "section_metadata/encyclopedia/material/idealized_structure/wyckoff_sets"
    },
    "idealized_structure": {
        "arch_source": "section_metadata/encyclopedia/material/idealized_structure"
    },
}

calculation_property_query = api.model("calculation_query", {
    "properties": fields.List(fields.String),
})
energies = api.model("energies", {
    "energy_total": fields.Float,
    "energy_total_T0": fields.Float,
    "energy_free": fields.Float,
})
electronic_band_structure = api.model("electronic_band_structure", {
    "reciprocal_cell": fields.List(fields.List(fields.Float)),
    "brillouin_zone": fields.Raw,
    "section_k_band_segment": fields.Raw,
    "section_band_gap": fields.Raw,
})
electronic_dos = api.model("electronic_dos", {
    "dos_energies": fields.List(fields.Float),
    "dos_values": fields.List(fields.List(fields.Float)),
})
calculation_property_result = api.model("calculation_property_result", {
    "lattice_parameters": fields.Nested(lattice_parameters, skip_none=True),
    "energies": fields.Nested(energies, skip_none=True),
    "mass_density": fields.Float,
    "atomic_density": fields.Float,
    "cell_volume": fields.Float,
    "wyckoff_sets": fields.Nested(wyckoff_set_result, skip_none=True),
    "idealized_structure": fields.Nested(idealized_structure_result, skip_none=True),
    "band_gap": fields.Float,
    "electronic_band_structure": fields.Nested(electronic_band_structure, skip_none=True),
    "electronic_dos": fields.Nested(electronic_dos, skip_none=True),
    "phonon_band_structure": fields.Raw,
    "phonon_dos": fields.Raw,
    "thermodynamical_properties": fields.Raw,
})


@ns.route("/materials/<string:material_id>/calculations/<string:calc_id>")
class EncCalculationResource(Resource):
    @api.response(404, "Material or calculation not found")
    @api.response(400, "Bad request")
    @api.response(200, "Metadata send", fields.Raw)
    @api.expect(calculation_property_query, validate=False)
    @api.marshal_with(calculation_property_result, skip_none=True)
    @api.doc("get_calculation")
    @authenticate()
    def post(self, material_id, calc_id):
        """Used to return calculation details. Some properties are not
        available in the ES index and are instead read from the Archive
        directly.
        """
        # Get query parameters as json
        try:
            data = marshal(request.get_json(), calculation_property_query)
        except Exception as e:
            abort(400, message=str(e))

        s = Search(index=config.elastic.index_name)
        query = Q(
            "bool",
            filter=get_enc_filter() + [
                Q("term", encyclopedia__material__material_id=material_id),
                Q("term", calc_id=calc_id),
            ]
        )
        s = s.query(query)

        # Create dictionaries for requested properties
        references = []
        properties = data["properties"]
        arch_properties = {}
        es_properties = {}
        ref_properties = set((
            "electronic_dos",
            "electronic_band_structure",
            "thermodynamical_properties",
            "phonon_dos",
            "phonon_band_structure",
        ))
        for prop in properties:
            es_source = calculation_property_map[prop].get("es_source")
            if es_source is not None:
                es_properties[prop] = es_source
                if prop in ref_properties:
                    references.append(prop)
            arch_source = calculation_property_map[prop].get("arch_source")
            if arch_source is not None:
                arch_properties[prop] = arch_source

        # The query is filtered already on the ES side so we don't need to
        # transfer so much data.
        sources = [
            "upload_id",
            "calc_id",
            "encyclopedia",
        ]
        sources += list(es_properties.values())

        s = s.extra(**{
            "_source": {"includes": sources},
            "size": 1,
        })

        response = s.execute()

        # No such material
        if len(response) == 0:
            abort(404, message="There is no material {} with calculation {}".format(material_id, calc_id))

        # Add references that are to be read from the archive
        for ref in references:
            arch_path = response[0]
            arch_path = rgetattr(arch_path, es_properties[ref])
            if arch_path is not None:
                arch_properties[ref] = arch_path
            del es_properties[ref]

        # If any of the requested properties require data from the Archive, the
        # file is opened and read.
        result = {}
        if len(arch_properties) != 0:
            entry = response[0]
            upload_id = entry.upload_id
            calc_id = entry.calc_id
            root = read_archive(
                upload_id,
                calc_id,
            )

            # Add results from archive
            for key, arch_path in arch_properties.items():
                value = root[arch_path]

                # Replace unnormalized thermodynamical properties with
                # normalized ones and turn into dict
                if key == "thermodynamical_properties":
                    specific_heat_capacity = value.specific_heat_capacity.magnitude.tolist()
                    specific_free_energy = value.specific_vibrational_free_energy_at_constant_volume.magnitude.tolist()
                    specific_heat_capacity = [x if np.isfinite(x) else None for x in specific_heat_capacity]
                    specific_free_energy = [x if np.isfinite(x) else None for x in specific_free_energy]
                if isinstance(value, list):
                    value = [x.m_to_dict() for x in value]
                else:
                    value = value.m_to_dict()
                if key == "thermodynamical_properties":
                    del value["thermodynamical_property_heat_capacity_C_v"]
                    del value["vibrational_free_energy_at_constant_volume"]
                    value["specific_heat_capacity"] = specific_heat_capacity
                    value["specific_vibrational_free_energy_at_constant_volume"] = specific_free_energy

                # DOS results are simplified.
                if key == "electronic_dos":
                    if "dos_energies_normalized" in value:
                        value["dos_energies"] = value["dos_energies_normalized"]
                        del value["dos_energies_normalized"]
                    if "dos_values_normalized" in value:
                        value["dos_values"] = value["dos_values_normalized"]
                        del value["dos_values_normalized"]

                # Pre-calculate k-path length to be used as x-coordinate in
                # plots. If the VBM and CBM information is needed later, it
                # can be added as indices along the path. The exact k-points
                # and occupations are removed to save some bandwidth.
                if key == "electronic_band_structure" or key == "phonon_band_structure":
                    segments = value["section_k_band_segment"]
                    k_path_length = 0
                    for segment in segments:
                        k_points = np.array(segment["band_k_points"])
                        segment_length = np.linalg.norm(k_points[-1, :] - k_points[0, :])
                        k_path_distances = k_path_length + np.linalg.norm(k_points - k_points[0, :], axis=1)
                        k_path_length += segment_length
                        segment["k_path_distances"] = k_path_distances.tolist()
                        del segment["band_k_points"]
                        if "band_occupations" in segment:
                            del segment["band_occupations"]

                result[key] = value

        # Add results from ES
        for prop, es_source in es_properties.items():
            value = rgetattr(response[0], es_source)
            if value is not None:
                if isinstance(value, AttrDict):
                    value = value.to_dict()
                result[prop] = value

        return result, 200


report_query = api.model("report_query", {
    "server": fields.String,
    "username": fields.String,
    "email": fields.String,
    "first_name": fields.String,
    "last_name": fields.String,
    "category": fields.String,
    "subcategory": fields.String(allow_null=True),
    "representatives": fields.Raw(Raw=True),
    "message": fields.String,
})


@ns.route("/materials/<string:material_id>/reports")
class ReportsResource(Resource):
    @api.response(500, "Error sending report")
    @api.response(400, "Bad request")
    @api.response(204, "Report succesfully sent", fields.Raw)
    @api.expect(calculation_property_query, validate=False)
    @api.marshal_with(calculation_property_result, skip_none=True)
    @api.doc("enc_report")
    @authenticate(required=True)
    def post(self, material_id):

        # Get query parameters as json
        try:
            query = marshal(request.get_json(), report_query)
        except Exception as e:
            abort(400, message=str(e))

        # Send the report as an email
        query["material_id"] = material_id
        representatives = query["representatives"]
        if representatives is not None:
            representatives = "\n" + "\n".join(["  {}: {}".format(key, value) for key, value in representatives.items()])
            query["representatives"] = representatives
        mail = (
            "Server: {server}\n\n"
            "Username: {username}\n"
            "First name: {first_name}\n"
            "Last name: {last_name}\n"
            "Email: {email}\n\n"
            "Material id: {material_id}\n"
            "Category: {category}\n"
            "Subcategory: {subcategory}\n"
            "Representative calculations: {representatives}\n\n"
            "Message: {message}"
        ).format(**query)
        try:
            infrastructure.send_mail(
                name="webmaster", email="lauri.himanen@gmail.com", message=mail, subject='Encyclopedia error report')
        except Exception as e:
            abort(500, message="Error sending error report email.")
        return "", 204


def read_archive(upload_id: str, calc_id: str) -> EntryArchive:
    """Used to read data from the archive.

    Args:
        upload_id: Upload id.
        calc_id: Calculation id.

    Returns:
        MSection: The section_run as MSection
        For each path, a dictionary containing the path as key and the returned
        section as value.
    """
    upload_files = UploadFiles.get(
        upload_id, is_authorized=create_authorization_predicate(upload_id, calc_id))

    with upload_files.read_archive(calc_id) as archive:
        data = archive[calc_id]
        root = EntryArchive.m_from_dict(data.to_dict())

    return root



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import Group

# Default
#
# All atoms belong to the “all” group.


class Delete(Group):
    name = 'delete'
    pass


class Region(Group):
    name = 'region'
    pass


class Type(Group):
    name = 'type'
    pass


class Id(Group):
    name = 'id'
    pass


class Molecule(Group):
    name = 'molecule'
    pass


class Variable(Group):
    name = 'variable'
    pass


class Include(Group):
    name = 'include'
    pass


class Subtract(Group):
    name = 'subtract'
    pass


class Union(Group):
    name = 'union'
    pass


class Intersect(Group):
    name = 'intersect'
    pass


class Dynamic(Group):
    name = 'dynamic'
    pass


class Static(Group):
    name = 'static'
    pass




import numpy as np
import logging

from ase.geometry import crystal_structure_from_cell
from ase.dft.kpoints import special_paths, parse_path_string
try:
    from ase.dft.kpoints import special_points
except ImportError:
    from ase.dft.kpoints import sc_special_points as special_points

from phonopy.phonon.band_structure import BandStructure
from phonopy.units import EvTokJmol, VaspToTHz


def generate_kpath_ase(cell, symprec):

    eig_val_max = np.real(np.linalg.eigvals(cell)).max()
    eps = eig_val_max * symprec

    lattice = crystal_structure_from_cell(cell, eps)
    paths = special_paths.get(lattice, None)
    if paths is None:
        paths = special_paths['orthorhombic']
    paths = parse_path_string(special_paths[lattice])
    points = special_points[lattice]
    k_points = []
    for p in paths:
        k_points.append([points[k] for k in p])
        for index in range(len(p)):
            if p[index] == 'G':
                p[index] = 'Γ'
    parameters = []
    for h, seg in enumerate(k_points):
        for i, path in enumerate(seg):
            parameter = {}
            parameter['npoints'] = 100
            parameter['startname'] = paths[h][i]
            if i == 0 and len(seg) > 2:
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
            elif i == (len(seg) - 2):
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
                break
            else:
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
    return parameters


class PhononProperties():
    def __init__(self, phonopy_obj, logger=None):
        if logger is None:
            logger = logging

        self.phonopy_obj = phonopy_obj

        self.n_atoms = len(phonopy_obj.unitcell)

        mesh_density = (2 * 80 ** 3) / self.n_atoms
        mesh_number = np.round(mesh_density**(1. / 3.))
        self.mesh = [mesh_number, mesh_number, mesh_number]

        self.n_atoms_supercell = len(phonopy_obj.supercell)

    def get_bandstructure(self):
        phonopy_obj = self.phonopy_obj

        frequency_unit_factor = VaspToTHz
        is_eigenvectors = False

        unit_cell = phonopy_obj.unitcell.get_cell()
        sym_tol = phonopy_obj.symmetry.tolerance
        parameters = generate_kpath_ase(unit_cell, sym_tol)

        # Distances calculated in phonopy.band_structure.BandStructure object
        # are based on absolute positions of q-points in reciprocal space
        # as calculated by using the cell which is handed over during instantiation.
        # Fooling that object by handing over a "unit cell" diag(1,1,1) instead clashes
        # with calculation of non-analytical terms.
        # Hence generate appropriate distances and special k-points list based on fractional
        # coordinates in reciprocal space (to keep backwards compatibility with previous
        # FHI-aims phonon implementation).
        bands = []
        bands_distances = []
        distance = 0.0
        bands_special_points = [distance]
        bands_labels = []
        label = parameters[0]["startname"]
        for b in parameters:
            kstart = np.array(b["kstart"])
            kend = np.array(b["kend"])
            npoints = b["npoints"]
            dk = (kend - kstart) / (npoints - 1)
            bands.append([(kstart + dk * n) for n in range(npoints)])
            dk_length = np.linalg.norm(dk)

            for n in range(npoints):
                bands_distances.append(distance + dk_length * n)

            distance += dk_length * (npoints - 1)
            bands_special_points.append(distance)
            label = [b["startname"], b["endname"]]
            bands_labels.append(label)

        bs_obj = BandStructure(
            bands, phonopy_obj.dynamical_matrix, with_eigenvectors=is_eigenvectors,
            factor=frequency_unit_factor)

        freqs = bs_obj.get_frequencies()

        return np.array(freqs), np.array(bands), np.array(bands_labels)

    def get_dos(self):
        phonopy_obj = self.phonopy_obj
        mesh = self.mesh

        phonopy_obj.set_mesh(mesh, is_gamma_center=True)
        q_points = phonopy_obj.get_mesh()[0]
        phonopy_obj.set_qpoints_phonon(q_points, is_eigenvectors=False)

        frequencies = phonopy_obj.get_qpoints_phonon()[0]
        self.frequencies = np.array(frequencies)
        min_freq = min(np.ravel(frequencies))
        max_freq = max(np.ravel(frequencies)) + max(np.ravel(frequencies)) * 0.05

        phonopy_obj.set_total_DOS(
            freq_min=min_freq, freq_max=max_freq, tetrahedron_method=True)
        f, dos = phonopy_obj.get_total_DOS()

        return f, dos

    def get_thermodynamical_properties(self):
        phonopy_obj = self.phonopy_obj

        t_max = 100
        t_min = 0
        t_step = 10
        mesh = self.mesh

        phonopy_obj.set_mesh(mesh, is_gamma_center=True)
        phonopy_obj.set_thermal_properties(t_step=t_step, t_max=t_max, t_min=t_min)
        T, fe, entropy, cv = phonopy_obj.get_thermal_properties()
        kJmolToEv = 1.0 / EvTokJmol
        fe = fe * kJmolToEv
        JmolToEv = kJmolToEv / 1000
        cv = JmolToEv * cv
        return T, fe, entropy, cv



import sys
from nomad.metainfo import Environment
from nomad.metainfo.legacy import LegacyMetainfoEnvironment
import siestaparser.metainfo.siesta
import nomad.datamodel.metainfo.common
import nomad.datamodel.metainfo.public
import nomad.datamodel.metainfo.general
import siestaparser.metainfo.siesta_autogenerated

m_env = LegacyMetainfoEnvironment()
m_env.m_add_sub_section(Environment.packages, sys.modules['siestaparser.metainfo.siesta'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.common'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.public'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.general'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['siestaparser.metainfo.siesta_autogenerated'].m_package)  # type: ignore



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='mopac_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='mopac.nomadmetainfo.json'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_mopac_fhof = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Final heat of formation
        ''',
        a_legacy=LegacyDefinition(name='x_mopac_fhof'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_mopac_keyword_line = Quantity(
        type=str,
        shape=[],
        description='''
        Mopac keyword line (it controls the calculation)
        ''',
        a_legacy=LegacyDefinition(name='x_mopac_keyword_line'))

    x_mopac_method = Quantity(
        type=str,
        shape=[],
        description='''
        Mopac method, i.e. PM7, AM1, etc..
        ''',
        a_legacy=LegacyDefinition(name='x_mopac_method'))


m_package.__init_metainfo__()



# Copyright 2016-2018 Fawzi Mohamed, Carl Poelking, Daria Tomecka
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from past.utils import old_div
from builtins import object
# See git https://github.com/capoe/momo.git
import os
import sys
import subprocess
import argparse
import time
import numpy as np
try:
	from lxml import etree
except ImportError:
    pass

boolean_dict = \
    {'true' : True,   '1' : True,  'yes' : True,
     'false' : False, '0' : False, 'no' : False, 'none' : False }

# =============================================================================
# XML WRAPPERS
# =============================================================================

class ExtendableNamespace(argparse.Namespace):
	def AddNamespace(self,  **kwargs):
		for name in kwargs:
			att = getattr(self, name, None)
			if att is None:
				setattr(self, name, kwargs[name])
			else:
				setattr(self, name, kwargs[name].As(type(att)))
		return
	def Add(self, name, value):
		att = getattr(self, name, None)
		if att is None:
			setattr(self, name, value)
		else:
			att.Add(name, value)
		return value

def GenerateTreeDict(tree, element, path='', paths_rel_to=None):
	if type(element) == etree._Comment: return [], {}
	# Update path
	if path == '':
		if element.tag != paths_rel_to:
			path += element.tag
	else:
		path += '/' + element.tag
	# Containers for lower levels
	tag_node = {}
	nodes = []
	# Construct Node
	xmlnode = XmlNode(element, path) # tree.getpath(element))
	nodes.append(xmlnode)
	if len(element) == 0:
		tag_node[path] = xmlnode
	#else:
	#	print "len 0", xmlnode.path
	#	tag_node[path] = xmlnode
	# Iterate over children
	for child in element:
		child_elements, childtag_element = GenerateTreeDict(tree, child, path)
		nodes = nodes + child_elements
		for key in childtag_element.keys():
			if key in tag_node:
				if type(tag_node[key]) != list:
					tag_node[key] = [ tag_node[key], childtag_element[key] ]
				else:
					tag_node[key].append(childtag_element[key])
			else:
				tag_node[key] = childtag_element[key]
	return nodes, tag_node

def NamespaceFromDict(tree_dict):
	nspace = ExtendableNamespace()
	for key in tree_dict.keys():
		#print "NSPACE Path = %s" % key
		sections = key.split('/')
		values = [ None for s in sections ]
		values[-1] = tree_dict[key]
		add_to_nspace = nspace
		for s,v in zip(sections, values):
			if v == None:
				#print "NSPACE Check for existing"
				if getattr(add_to_nspace, s, None):
					#print "NSPACE '%s' already exists" % s
					add_to_nspace = getattr(add_to_nspace, s, None)
				else:
					#print "NSPACE '%s' created" % s
					sub_nspace = ExtendableNamespace()
					add_to_nspace = add_to_nspace.Add(s, sub_nspace)
			else:
				#print "NSPACE '%s' value added" % s
				add_to_nspace.Add(s, v)
	return nspace


class XmlTree(list):
	def __init__(self, xmlfile, paths_rel_to=None):
		self.xmlfile = xmlfile
		self.xtree = etree.parse(xmlfile)
		self.xroot = self.xtree.getroot()
		self.nodes, self.tag_node = GenerateTreeDict(self.xtree, self.xroot, '', paths_rel_to)
		self.xspace = NamespaceFromDict(self.tag_node)
	def SelectByTag(self, tag):
		selection = [ e for e in self.nodes if e.tag == tag ]
		return selection
	def __getitem__(self, key):
		return self.tag_node[key]
	def keys(self):
		return list(self.tag_node.keys())


class XmlNode(object):
	def __init__(self, element, path):
		self.path = path
		self.node = element
		self.tag = element.tag
		self.value = element.text
		self.attributes = element.attrib
	def As(self, typ):
		if typ == np.array:
			sps = self.value.split()
			return typ([ float(sp) for sp in sps ])
		elif typ == bool:
			return boolean_dict.get(self.value.lower())
		else:
			return typ(self.value)
	def AsArray(self, typ, sep=' ', rep='\t\n'):
		for r in rep:
			self.value = self.value.replace(r, sep)
		sp = self.value.split(sep)
		return [ typ(s) for s in sp if str(s) != '' ]
	def SetNodeValue(self, new_value):
		self.value = new_value
		if self.node != None:
			self.node.firstChild.nodeValue = new_value
		return
	def __getitem__(self, key):
		return self.node.get(key)


# =============================================================================
# COMMAND LINE & XML INPUT INTERFACE
# =============================================================================

class CLIO_HelpFormatter(argparse.HelpFormatter):
	def _format_usage(self, usage, action, group, prefix):
		return "%s : Command Line Interface\n" % sys.argv[0]


class OptionsInterface(object):
	def __init__(self):
		# COMMAND-LINE ARGUMENTS
		self.is_connected_to_cmd_ln = False
		self.cmd_ln_args = None
		self.cmd_ln_opts = None
		self.cmd_ln_nicknames = [ '-h' ]
		self.boolean_translator = boolean_dict
		self.subtype = str
		# XML OPTIONS FILE
		self.is_connected_to_xml = False
		self.xmlfile = None
		self.tree = None
		self.xdict = None
		self.xspace = None
		# JOINED OPTIONS
		self.opts = ExtendableNamespace()
	def Connect(self, xmlfile=None):
		self.ConnectToCmdLn()
		self.ConnectToOptionsFile(xmlfile)
	def Parse(self, xkey='options'):
		if self.is_connected_to_cmd_ln:
			self.ParseCmdLn()
		if self.is_connected_to_xml:
			self.ParseOptionsFileXml(xkey)
		if self.is_connected_to_cmd_ln and not self.is_connected_to_xml:
			return self.cmd_ln_opts
		elif self.is_connected_to_xml and not self.is_connected_to_cmd_ln:
			return self.xspace
		else:
			return self.cmd_ln_opts, self.xspace
	def ParseOptionsFile(self, xmlfile, xkey):
		self.xmlfile = xmlfile
		self.is_connected_to_xml = True
		self.ParseOptionsFileXml(xkey)
		return self.xspace
	# COMMAND-LINE CONVENIENCE ================================================
	def __call__(self):
		return self.cmd_ln_opts
	def ConnectToCmdLn(self, prog=sys.argv[0], descr=None):
		self.cmd_ln_args = argparse.ArgumentParser(prog=sys.argv[0],
			formatter_class=lambda prog: CLIO_HelpFormatter(prog,max_help_position=70))
		self.is_connected_to_cmd_ln = True
		return
	def ParseCmdLn(self):
		self.cmd_ln_opts = self.cmd_ln_args.parse_args()
	def InterpretAsBoolean(self, expr):
		try:
			return self.boolean_translator.get(expr.lower())
		except KeyError:
			raise ValueError('CLIO does not know how to convert %s into a boolean.' % expr)
	def InterpretAsNumpyArray(self, expr):
		# print("Interpret", expr)
		array = [ float(e) for e in expr ]
		array = np.array(array)
		return array
	def InterpretAsList(self, expr):
		array = [ self.subtype(e) for e in expr ]
		return array
	def AddArg(self, name, typ=str, nickname=None,
			default=None, destination=None, help=None):
		# Sort out <name> (e.g. --time) vs <destination> (e.g., time)
		if '--' != name[0:2]:
			dest = name
			name = '--' + name
		else:
			dest = name[2:]
		# Sort out <default> vs <required>
		if default == None: required = True
		else: required = False
		# Construct default <help> it not given
		if help == None: help = str(typ)
		# Construct <nickname> if not given
		if nickname == None:
			nickname = '-'
			for char in dest:
				nickname += char
				if not nickname in self.cmd_ln_nicknames:
					break
			if nickname in self.cmd_ln_nicknames:
				raise ValueError('CLIO could not construct nickname from %s option'\
					% name)
			self.cmd_ln_nicknames.append(nickname)
		# Process type
		if typ in [int, float, str]:
			nargs = None
		elif typ == bool:
			typ = self.InterpretAsBoolean
			nargs = None
		elif typ == np.array:
			raise NotImplementedError
			typ = float # self.InterpretAsNumpyArray
			nargs = 3
		elif typ == list:
			typ = str
			nargs = '*'
		elif len(typ) == 2 and typ[0] == list:
			typ = typ[1]
			nargs = '*'
		else:
			raise NotImplementedError("CLIO does not know how to generate type '%s'"\
				 % typ)
		self.cmd_ln_args.add_argument(nickname, name,
				dest=dest,
				action='store',
				nargs=nargs,
				required=required,
				type=typ,
				metavar=dest[0:1].upper(),
				default=default,
				help=help)
		return
	# COMMAND-LINE CONVENIENCE ================================================
	def ConnectToOptionsFile(self, xmlfile):
		if xmlfile == None or xmlfile == '': return
		self.xmlfile = xmlfile
		self.is_connected_to_xml = True
		return
	def ParseOptionsFileXml(self, xkey='options'):
		if self.xmlfile == None: return
		self.tree = XmlTree(self.xmlfile, paths_rel_to=xkey)
		self.xdict = self.tree.tag_node
		self.xspace = self.tree.xspace
		return
	def __getitem__(self, key):
		try:
			return self.xspace.__dict__[key]
		except KeyError:
			return self.cmd_ln_opts.__dict__[key]
		except KeyError:
			raise AttributeError('No such option registered: \'%s\'' % key)
		return None


# =============================================================================
# OS SHELL INTERFACE
# =============================================================================

class ShellInterface(object):
	def __init__(self):
		# =====================================================================
		# PRINTER ATTRIBUTES
		self.color_dict = { \
			'pp' : '\033[95m',
			'mb' : '\033[34m',
			'lb' : '\033[1;34m',
			'my' : '\033[1;33m',
			'mg' : '\033[92m',
			'mr' : '\033[91m',
			'ww' : '\033[0;1m',
			'ok' : '\033[92m',
			'xx' : '\033[91m',
			'warning' : '\033[93m',
			'error' : '\033[95m',
			'endcolor' : '\033[0;1m' }
		self.justify_dict = { \
			'o' : '  o ',
			'.' : '... ',
            'r' : '\r',
			'ro' : '\r  o '}
		self.pp = OS_COLOR('pp')
		self.lb = OS_COLOR('lb')
		self.mb = OS_COLOR('mb')
		self.mg = OS_COLOR('mg')
		self.my = OS_COLOR('my')
		self.mr = OS_COLOR('mr')
		self.ww = OS_COLOR('ww')
		self.ok = OS_COLOR('ok')
		self.xx = OS_COLOR('xx')
		self.colors = [ OS_COLOR(c) for c in sorted(self.color_dict.keys()) ]
		self.item = '  o '
		self.iitem = '      - '
		self.endl = OS_LINE_CHAR('\n')
		self.flush = OS_LINE_CHAR('')
		self.back = OS_LINE_CHAR('\r')
		self.trail = ' '
		# CURRENT STYLE SELECTION
		self.sel_color = None
		self.sel_justify = None
		self.sel_header = False
		self.sel_trim = "="
		# EXE ATTRIBUTES
		self.catch = OS_EXE_CATCH()
		self.assert_zero = OS_EXE_ASSERT()
		self.dev = OS_EXE_DEV('')
		self.nodev = OS_EXE_DEV('')
		self.devnull = OS_EXE_DEV()
		self.devfile = OS_EXE_DEV('')
		self.os_exe_get = False
		self.os_exe_assert_zero = False
		self.os_exe_dev = ''
		self.os_exe_verbose = False
		# LOGGING
		self.logfile = None
		# DIRECTORY HOPPING
		self.paths_visited = [ os.getcwd() ]
		self.exe_root_path = self.paths_visited[0]
		self.N_store_paths_visited = 1+5
	def __call__(self, mssg, c=None, j=None, h=False, t="="):
		# c=color, j=justify, h=header, t=trim, u=upper-case
		if j:
			mssg = self.justify_dict[j] + mssg
		if c != None:
			mssg = self.color_dict[c] + mssg + self.color_dict['endcolor']
		if h:
			mssg = self.os_generate_header(mssg, t)
		# print(mssg)
	# LOGFILE ADAPTOR =========================================================
	def ConnectToFile(self, logfile):
		self.logfile = logfile
		sys.stdout = open(logfile, 'w')
		self.devfile = OS_EXE_DEV(' >> {log} 2>> {log}'.format(log=logfile))
		return
	def DisconnectFromFile(self):
		if self.logfile != None:
			self.devfile = OS_EXE_DEV('')
			self.logfile = None
			sys.stdout = sys.__stdout__
		else: pass
		return
	# PRINTER METHODS =========================================================
	def __lshift__(self, mssg):
		if type(mssg) == OS_LINE_CHAR:
			# <LOG MESSAGE HERE>
			sys.stdout.write(str(mssg))
			sys.stdout.flush()
			self.sel_color = None
			return self
		elif type(mssg) == OS_COLOR:
			self.sel_color = str(mssg)
			return self
		mssg = str(mssg)
		if self.sel_justify != None:
			mssg = self.justify_dict[j] + mssg
		mssg += self.trail
		if self.sel_color != None:
			mssg = self.color_dict[self.sel_color] \
				+ mssg + self.color_dict['endcolor']
		if self.sel_header:
			mssg = self.os_generate_header(mssg, self.sel_trim)
		# <LOG MESSAGE HERE>
		sys.stdout.write(mssg)
		return self
	def os_print(self, mssg, c=None, j=None, h=False, t="="):
		# c=color, j=justify, h=header, t=trim, u=upper-case
		if j:
			mssg = OS_JUSTIFY_DICT[j] + mssg
		if c != None:
			mssg = OS_COLOR_DICT[c] + mssg + OS_COLOR_DICT['endcolor']
		if h:
			mssg = os_generate_header(mssg, t)
		# print(mssg)
		return
	def os_print_config(self, c=None, j=None, h=False, t="=", tl=' '):
		self.sel_color = c
		self.sel_justify = j
		self.sel_header = h
		self.sel_trim = t
		self.trail = tl
		return
	def os_print_reset(self):
		self.sel_color = None
		self.sel_justify = None
		self.sel_header = False
		self.sel_trim = "="
		self.trail = ' '
		return
	def os_generate_header(title, trim="="):
		try:
		    height, width = os.popen('stty size', 'r').read().split()
		    width = int(width)
		    leftright = ((width - len(title)-2)) // 2
		except ValueError:
		    leftright = 40
		return trim*leftright + " " + title + " " + trim*leftright
	# EXECUTE METHODS =========================================================
	def __rshift__(self, cmmd):
		if type(cmmd) == OS_EXE_CATCH:
			self.os_exe_get = True
			return self
		elif type(cmmd) == OS_EXE_DEV:
			self.dev = cmmd
			return self
		elif type(cmmd) == OS_EXE_ASSERT:
			self.os_exe_assert_zero = True
			return self
		# Redirect command as requested (not applicable if output is asked for)
		if not self.os_exe_get:
			if str(self.dev) != '':
				cmmd += str(self.dev)
				self.dev = self.nodev
			else:
				cmmd += str(self.devfile)
		# Execute
		if self.debug: self << self.my << "exe:" << cmmd << endl
		if self.os_exe_get:
			output = subprocess.getoutput(cmmd)
			self.os_exe_get = False
			return output
		else:
			sign = os.system(cmmd)
		if self.os_exe_assert_zero:
			if str(sign) != '0':
				raise RuntimeError("<OSIO> '%s' returned '%s'" % (cmmd, sign))
			self.os_exe_assert_zero = False
		return sign
	# EXIT PROTOCOLS ==========================================================
	def okquit(self, what=''):
		if what != '': self << self.ok << what << self.endl
		self.DisconnectFromFile()
		sys.exit(0)
		return
	def xxquit(self, what=''):
		if what != '': self << self.xx << "ERROR" << what << self.endl
		self.DisconnectFromFile()
		sys.exit(1)
		return
	# DIRECTORY HOPPING =======================================================
	def cd(self, d):
		# Current working directory, for archiving ... =>
		cwd = os.getcwd()
		if type(d) == int:
			# Change to previously visited path
			os.chdir(self.paths_visited[d])
		elif type(d) == str:
			# Change to path as specified explicitly
			os.chdir(d)
		else:
			raise NotImplementedError
		# <= ... previous path
		self.paths_visited.append(cwd)
		if len(self.paths_visited) > self.N_store_paths_visited:
			self.paths_visited.pop(1) # 0 stores root
		if self.debug: self << self.my << "cd: " << os.getcwd() << self.endl
		return
	def pwd(self):
		return self.cwd()
	def cwd(self):
		return os.getcwd()
	def root(self):
		self.cd(self.exe_root_path)
		return
	def abspath(self, file):
		if not os.path.exists(file):
			raise IOError("<osio::abspath> No such item in local directory: '%s'" % file)
		return os.path.join(self.cwd(), file)
	def mkcd(self, directory):
		self >> self.assert_zero >> 'mkdir -p %s' % directory
		self.cd(directory)
		return directory

class OS_EXE_DEV(object):
	def __init__(self, dev=' > /dev/null 2> /dev/null'):
		self.dev = dev
	def __str__(self):
		return self.dev

class OS_EXE_CATCH(object):
	def __init__(self):
		self.catch = True

class OS_EXE_ASSERT(object):
	def __init__(self):
		self.assert_0 = True

class OS_COLOR(object):
	def __init__(self, colstr):
		self.colstr = colstr
	def __str__(self):
		return self.colstr

class OS_LINE_CHAR(object):
	def __init__(self, char):
		self.char = char
	def __str__(self):
		return self.char


class OSIO(ShellInterface, OptionsInterface):
	def __init__(self):
		self.debug = False
		ShellInterface.__init__(self)
		OptionsInterface.__init__(self)
		return
	def sleep(self, dt):
		time.sleep(dt)
		return


osio = OSIO()
endl = OS_LINE_CHAR('\n')
flush = OS_LINE_CHAR('')
back = OS_LINE_CHAR('\r')
catch = OS_EXE_CATCH()
devnull = OS_EXE_DEV()




# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

functionals = {'LSDA': ['LDA_X', 'LDA_C_VWN_3'],
               'LDA': ['LDA_X', 'LDA_C_VWN_3'],
               'SVWN': ['LDA_X', 'LDA_C_VWN_3'],
               'LSDA5': ['LDA_X', 'LDA_C_VWN'],
               'LDA5': ['LDA_X', 'LDA_C_VWN'],
               'SVWN5': ['LDA_X', 'LDA_C_VWN'],
               'HFB': ['GGA_X_B88'], #['GGA_X_B88' + Becke's 1988 exchange functional which includes the Slater exchange along with corrections involving the gradient of the density],
               'HFS': ['LDA_X'],  # Original HK/KS articles?  No libxc name...
               'HFB86': ['GGA_X_B86'],
               'HFO': ['GGA_X_OPTX'],
               'BLYP': ['GGA_X_B88', 'GGA_C_LYP'],
               'BPBE': ['GGA_X_B88', 'GGA_C_PBE'],
               'B3LYP': ['HYB_GGA_XC_B3LYP'],
               'B3LYP5': ['HYB_GGA_XC_B3LYP5'],
               #'B2PLYP': {'LDA_X': 0.47,
               #           'HF_X': 1. - 0.47,
               #           'GGA_X_B88': 0.47,
               #           },  # No clue how to set this one
               'B86LYP': ['GGA_X_B86', 'GGA_C_LYP'],
               'BWIG': ['GGA_X_B88', 'LDA_C_WIGNER'],
               'GLYP': ['GGA_X_G96', 'GGA_C_LYP'],
               'OLYP': ['GGA_X_OPT', 'GGA_C_LYP'],
               'OPBE': ['GGA_X_OPT', 'GGA_C_PBE'],
               'O3LYP': ['HYB_GGA_XC_O3LYP'],  # Or hideous composite thingy?
               #'O3LYP': {'HF_X': 0.1161,
               #          'LDA_X': 0.9262,
               #          'GGA_X_OPT': 0.8133,
               #          'GGA_C_VWN': 0.19,
               #          'GGA_C_LYP': 0.81}, ?
               #'KT3': ['GGA_X_KT...3'], TODO add to nomad xc functionals
               'TLYP': ['HF_X', 'GGA_C_LYP'],
               'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
               'PBE0': None,  # wtf
               'PBESOL': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
               'RGE2': ['GGA_X_RGE2', 'GGA_C_PBE_SOL'],
               'PTCA': ['GGA_X_PBE', 'GGA_C_TCA'],
               'SSB': ['GGA_X_SSB', 'GGA_C_PBE'],
               'M06': ['HYB_MGGA_XC_M06'],
               'M06L': ['MGGA_C_M06_L', 'MGGA_X_M06_L'],
               'M06HF': ['HYB_MGGA_XC_M06_HF'],  # Is this good enough?
               'M062X': ['HYB_MGGA_XC_M06_2X']}



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
The raw API of the nomad@FAIRDI APIs. Can be used to retrieve raw calculation files.
'''

from typing import IO, Any, Union, List
import os.path
from io import BytesIO
from flask import request, send_file
from flask_restplus import abort, Resource, fields
import magic
import fnmatch
import json
import gzip
import lzma
import urllib.parse

from nomad import search, utils, config
from nomad.files import UploadFiles, Restricted
from nomad.processing import Calc
from nomad.app import common

from .api import api
from .auth import authenticate, create_authorization_predicate
from .common import streamed_zipfile, add_search_parameters, apply_search_parameters,\
    search_model


ns = api.namespace('raw', description='Downloading raw data files.')

raw_file_list_model = api.model('RawFileList', {
    'upload_id': fields.String(description='The id of the upload.'),
    'directory': fields.String(description='The path to the directory in the upload.'),
    'contents': fields.List(
        fields.Nested(model=api.model('RawFileListContents', {
            'file': fields.String(description='The file name'),
            'size': fields.Integer(description='The file size in bytes')
        })))})

raw_file_compress_argument = dict(
    name='compress', type=bool, help='Use compression on .zip files, default is not.',
    location='args')
raw_file_strip_argument = dict(
    name='strip', type=bool, help='Removes a potential common path prefix from all file paths.',
    location='args')

_raw_file_from_path_parser = api.parser()
_raw_file_from_path_parser.add_argument(**raw_file_compress_argument)
_raw_file_from_path_parser.add_argument(**raw_file_strip_argument)
_raw_file_from_path_parser.add_argument(
    name='length', type=int, help='Download only x bytes from the given file.',
    location='args')
_raw_file_from_path_parser.add_argument(
    name='offset', type=int, help='Start downloading a file\' content from the given offset.',
    location='args')
_raw_file_from_path_parser.add_argument(
    name='decompress', type=int, help='Automatically decompress the file if compressed. Only supports .gz',
    location='args')


class FileView:
    '''
    File-like wrapper that restricts the contents to a portion of the file.
    Arguments:
        f: the file-like
        offset: the offset
        length: the amount of bytes
    '''
    def __init__(self, f, offset, length):
        self.f = f
        self.f_offset = offset
        self.offset = 0
        self.length = length

    def seek(self, offset, whence=0):
        if whence == os.SEEK_SET:
            self.offset = offset
        elif whence == os.SEEK_CUR:
            self.offset += offset
        elif whence == os.SEEK_END:
            self.offset = self.length + offset
        else:
            # Other values of whence should raise an IOError
            return self.f.seek(offset, whence)
        return self.f.seek(self.offset + self.f_offset, os.SEEK_SET)

    def tell(self):
        return self.offset

    def read(self, size=-1):
        self.seek(self.offset)
        if size < 0:
            size = self.length - self.offset
        size = max(0, min(size, self.length - self.offset))
        self.offset += size
        return self.f.read(size)


def get_raw_file_from_upload_path(
        upload_files, upload_filepath, authorization_predicate, mainfile: str = None):
    '''
    Helper method used by func:`RawFileFromUploadPathResource.get` and
    func:`RawFileFromCalcPathResource.get`.
    '''
    upload_filepath = upload_filepath.rstrip('/')

    if upload_filepath[-1:] == '*':
        upload_filepath = upload_filepath[0:-1]
        wildcarded_files = list(upload_files.raw_file_manifest(path_prefix=upload_filepath))
        if len(wildcarded_files) == 0:
            abort(404, message='There are no files for %s.' % upload_filepath)
        else:
            compress = request.args.get('compress', None) is not None
            strip = request.args.get('strip', None) is not None
            return respond_to_get_raw_files(
                upload_files.upload_id, wildcarded_files, compress=compress, strip=strip)

    try:
        try:
            offset = int(request.args.get('offset', 0))
            length = int(request.args.get('length', -1))
        except Exception:
            abort(400, message='bad parameter types')

        if offset < 0:
            abort(400, message='bad offset, length values')
        if offset > 0 and length <= 0:
            abort(400, message='bad offset, length values')

        raw_file = upload_files.raw_file(upload_filepath, 'br')

        if request.args.get('decompress') is not None:
            if upload_filepath.endswith('.gz'):
                filename = upload_filepath[:3]
                upload_filepath = filename
                raw_file = gzip.GzipFile(filename=filename, mode='rb', fileobj=raw_file)
            if upload_filepath.endswith('.xz'):
                filename = upload_filepath[:3]
                upload_filepath = filename
                raw_file = lzma.open(filename=raw_file, mode='rb')

        buffer = raw_file.read(2048)
        raw_file.seek(0)
        mime_type = magic.from_buffer(buffer, mime=True)

        raw_file_view: Union[FileView, IO[Any]] = None
        if length > 0:
            raw_file_view = FileView(raw_file, offset, length)
        else:
            raw_file_view = raw_file

        return send_file(
            raw_file_view,
            mimetype=mime_type,
            as_attachment=True,
            cache_timeout=0,
            attachment_filename=os.path.basename(upload_filepath))
    except Restricted:
        abort(401, message='Not authorized to access all files in %s.' % upload_files.upload_id)
    except KeyError:
        directory_files = upload_files.raw_file_list(upload_filepath)
        if len(directory_files) == 0:
            abort(404, message='There is nothing to be found at %s.' % upload_filepath)

        contents = sorted([dict(name=name, size=size) for name, size in directory_files], key=lambda x: '' if x['name'] == mainfile else x['name'])
        # if mainfile is not None:
        #     contents = [mainfile] + [content for content in contents if content['name'] != mainfile]
        return {
            'upload_id': upload_files.upload_id,
            'directory': upload_filepath,
            'contents': contents
        }, 200


@ns.route('/<string:upload_id>/<path:path>')
@api.doc(params={
    'upload_id': 'The unique id for the requested upload.',
    'path': 'The path to a file or directory with optional wildcard.'
})
class RawFileFromUploadPathResource(Resource):
    @api.doc('get')
    @api.response(404, 'The upload or path does not exist')
    @api.response(401, 'Not authorized to access the requested files.')
    @api.response(200, 'File(s) send')
    @api.expect(_raw_file_from_path_parser, validate=True)
    @authenticate(signature_token=True)
    def get(self, upload_id: str, path: str):
        ''' Get a single raw calculation file, directory contents, or whole directory sub-tree
        from a given upload.

        The 'upload_id' parameter needs to identify an existing upload.

        If the upload
        is not yet published or contains requested data with embargo, proper authentication
        is required. This can be done via HTTP headers as usual. But, if you need to
        access files via plain URLs (e.g. for curl, download link, etc.), URLs for
        this endpoint can be token signed (see also /auth/token). For unpublished
        uploads, authentication is required regardless. For (partially) embargoed data,
        multi file downloads work, but will not contain any embargoed data.

        If the given path points to a file, the file is provided with the appropriate
        Content-Type header. A 401 is returned for staging, embargo files with unsigned
        or wrongly signed URLs. When accessing a file, the additional query parameters 'length'
        and 'offset' can be used to partially download a file's content.

        If the given path points to a directory, the content (names, sizes, type) is returned
        as a json body. Only visible items (depending on authenticated user, token) are
        returned.

        If the given path ends with the '*' wildcard character, all upload contents that
        match the given path at the start, will be returned as a .zip file body.
        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.
        '''
        # TODO this is a quick fix, since swagger cannot deal with not encoded path parameters
        if path is not None:
            path = urllib.parse.unquote(path)

        upload_filepath = path

        # TODO find a better way to all access to certain files
        if os.path.basename(path).endswith('.png'):
            def authorization_predicate(*args, **kwargs):
                return True
        else:
            authorization_predicate = create_authorization_predicate(upload_id)

        upload_files = UploadFiles.get(upload_id, authorization_predicate)
        if upload_files is None:
            abort(404, message='The upload with id %s does not exist.' % upload_id)

        return get_raw_file_from_upload_path(upload_files, upload_filepath, authorization_predicate)


@ns.route('/calc/<string:upload_id>/<string:calc_id>/<path:path>')
@api.doc(params={
    'upload_id': 'The unique id for the requested calc\'s upload.',
    'calc_id': 'The unique calc id for the requested calc',
    'path': 'The path to a file or directory with optional wildcard.'
})
class RawFileFromCalcPathResource(Resource):
    @api.doc('get_file_from_calc')
    @api.response(404, 'The upload or path does not exist')
    @api.response(401, 'Not authorized to access the requested files.')
    @api.response(200, 'File(s) send')
    @api.expect(_raw_file_from_path_parser, validate=True)
    @authenticate(signature_token=True)
    def get(self, upload_id: str, calc_id: str, path: str):
        ''' Get a single raw calculation file, calculation contents, or all files for a
        given calculation.

        The 'upload_id' parameter needs to identify an existing upload.
        The 'calc_id' parameter needs to identify a calculation within in the upload.

        This endpoint behaves exactly like /raw/<upload_id>/<path>, but the path is
        now relative to the calculation and not the upload.
        '''
        # TODO this is a quick fix, since swagger cannot deal with not encoded path parameters
        if path is not None:
            path = urllib.parse.unquote(path)

        calc_filepath = path if path is not None else ''
        authorization_predicate = create_authorization_predicate(upload_id, calc_id=calc_id)
        upload_files = UploadFiles.get(upload_id, authorization_predicate)
        if upload_files is None:
            abort(404, message='The upload with id %s does not exist.' % upload_id)

        try:
            calc = Calc.get(calc_id)
        except KeyError:
            abort(404, message='The calc with id %s does not exist.' % calc_id)

        if calc.upload_id != upload_id:
            abort(404, message='The calc with id %s is not part of the upload with id %s.' % (calc_id, upload_id))

        upload_filepath = os.path.join(os.path.dirname(calc.mainfile), calc_filepath)
        return get_raw_file_from_upload_path(
            upload_files, upload_filepath, authorization_predicate,
            mainfile=os.path.basename(calc.mainfile))


@ns.route('/calc/<string:upload_id>/<string:calc_id>/')
class RawFileFromCalcEmptyPathResource(RawFileFromCalcPathResource):
    @api.doc('get_file_list_from_calc')
    @api.response(404, 'The upload or path does not exist')
    @api.response(401, 'Not authorized to access the requested files.')
    @api.response(200, 'File(s) send')
    @api.expect(_raw_file_from_path_parser, validate=True)
    @authenticate(signature_token=True)
    def get(self, upload_id: str, calc_id: str):
        ''' Get calculation contents.

        This is basically /raw/calc/<upload_id>/<calc_id>/<path> with an empty path, since
        having an empty path parameter is not possible.
        '''
        return super().get(upload_id, calc_id, None)


_raw_files_request_model = api.model('RawFilesRequest', {
    'files': fields.List(
        fields.String, default=[], description='List of files to download.'),
    'compress': fields.Boolean(
        default=False,
        description='Enable compression, default is not compression.')
})

_raw_files_request_parser = api.parser()
_raw_files_request_parser.add_argument(
    'files', required=True, type=str, help='Comma separated list of files to download.', location='args')
_raw_files_request_parser.add_argument(
    'prefix', type=str, help='A common prefix that is prepend to all files', location='args')
_raw_files_request_parser.add_argument(**raw_file_strip_argument)
_raw_files_request_parser.add_argument(**raw_file_compress_argument)


@ns.route('/<string:upload_id>')
@api.doc(params={
    'upload_id': 'The unique id for the requested upload.'
})
class RawFilesResource(Resource):
    @api.doc('get_files')
    @api.response(404, 'The upload or path does not exist')
    @api.response(200, 'File(s) send', headers={'Content-Type': 'application/zip'})
    @api.expect(_raw_files_request_model, validate=True)
    @authenticate()
    def post(self, upload_id):
        ''' Download multiple raw calculation files in a .zip file.

        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.
        '''
        json_data = request.get_json()
        compress = json_data.get('compress', False)
        files = [file.strip() for file in json_data['files']]

        return respond_to_get_raw_files(upload_id, files, compress)

    @api.doc('get_files_alternate')
    @api.response(404, 'The upload or path does not exist')
    @api.response(200, 'File(s) send', headers={'Content-Type': 'application/zip'})
    @api.expect(_raw_files_request_parser, validate=True)
    @authenticate(signature_token=True)
    def get(self, upload_id):
        '''
        Download multiple raw calculation files.
        Download multiple raw calculation files in a .zip file.
        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.
        '''
        args = _raw_files_request_parser.parse_args()

        files_str = args.get('files')
        if files_str is None:
            abort(400, message="No files argument given.")

        prefix = args.get('prefix')
        if prefix is not None:
            files = [os.path.join(prefix, file.strip()) for file in files_str.split(',')]
        else:
            files = [file.strip() for file in files_str.split(',')]

        compress = args.get('compress', False)
        strip = args.get('strip', False)

        return respond_to_get_raw_files(upload_id, files, compress=compress, strip=strip)


_raw_file_from_query_parser = api.parser()
add_search_parameters(_raw_file_from_query_parser)
_raw_file_from_query_parser.add_argument(
    name='compress', type=bool, help='Use compression on .zip files, default is not.',
    location='args')
_raw_file_from_query_parser.add_argument(**raw_file_strip_argument)
_raw_file_from_query_parser.add_argument(
    name='file_pattern', type=str,
    help=(
        'A wildcard pattern. Only filenames that match this pattern will be in the '
        'download. Multiple patterns will be combined with logical or'),
    location='args', action='append')


_raw_file_from_query_model = api.inherit('ArchiveSearch', search_model, {
    'compress': fields.Boolean(description='Use compression on .zip files, default is not.', default=False),
    'strip': fields.Boolean(description='Removes a potential common path prefix from all file paths.', default=False),
    'file_pattern': fields.List(fields.String, description=(
        'A wildcard pattern. Only filenames that match this pattern will be in the '
        'download. Multiple patterns will be combined with logical or'), allow_null=True, skip_none=True)
})


@ns.route('/query')
class RawFileQueryResource(Resource):
    manifest_quantities = ['upload_id', 'calc_id', 'external_id', 'raw_id', 'pid', 'calc_hash']

    @api.doc('raw_files_from_query')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.expect(_raw_file_from_query_parser, validate=True)
    @api.response(200, 'File(s) send', headers={'Content-Type': 'application/zip'})
    @authenticate(signature_token=True)
    def get(self):
        ''' Download a .zip file with all raw-files for all entries that match the given
        search parameters.

        See ``/repo`` endpoint for documentation on the search
        parameters.

        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.

        The zip file will contain a ``manifest.json`` with the repository meta data.
        '''
        logger = common.logger.bind(query=urllib.parse.urlencode(request.args, doseq=True))

        patterns: List[str] = None
        try:
            args = _raw_file_from_query_parser.parse_args()
            compress = args.get('compress', False)
            strip = args.get('strip', False)
            pattern = args.get('file_pattern', None)
            if isinstance(pattern, str):
                patterns = [pattern]
            elif pattern is None:
                patterns = []
            else:
                patterns = pattern
        except Exception:
            abort(400, message='bad parameter types')

        search_request = search.SearchRequest()
        apply_search_parameters(search_request, _raw_file_from_query_parser.parse_args())
        search_request.include('calc_id', 'upload_id', 'mainfile')

        def path(entry):
            return '%s/%s' % (entry['upload_id'], entry['mainfile'])

        calcs = search_request.execute_scan(
            order_by='upload_id',
            size=config.services.download_scan_size,
            scroll=config.services.download_scan_timeout)

        if strip:
            if search_request.execute()['total'] > config.raw_file_strip_cutoff:
                abort(400, 'The requested download has to many files for using "strip".')
            calcs = list(calcs)
            paths = [path(entry) for entry in calcs]
            common_prefix_len = len(utils.common_prefix(paths))
        else:
            common_prefix_len = 0

        def generator():
            try:
                manifest = {}
                directories = set()
                upload_files = None
                streamed, skipped = 0, 0

                for entry in calcs:
                    upload_id = entry['upload_id']
                    mainfile = entry['mainfile']
                    if upload_files is None or upload_files.upload_id != upload_id:
                        logger.info('opening next upload for raw file streaming', upload_id=upload_id)
                        if upload_files is not None:
                            upload_files.close()

                        upload_files = UploadFiles.get(upload_id)

                        if upload_files is None:
                            logger.error('upload files do not exist', upload_id=upload_id)
                            continue

                        def open_file(upload_filename):
                            return upload_files.raw_file(upload_filename, 'rb')

                    upload_files._is_authorized = create_authorization_predicate(
                        upload_id=upload_id, calc_id=entry['calc_id'])
                    directory = os.path.dirname(mainfile)
                    directory_w_upload = os.path.join(upload_files.upload_id, directory)
                    if directory_w_upload not in directories:
                        streamed += 1
                        directories.add(directory_w_upload)
                        for filename, file_size in upload_files.raw_file_list(directory=directory):
                            filename = os.path.join(directory, filename)
                            filename_w_upload = os.path.join(upload_files.upload_id, filename)
                            filename_wo_prefix = filename_w_upload[common_prefix_len:]
                            if len(patterns) == 0 or any(
                                    fnmatch.fnmatchcase(os.path.basename(filename_wo_prefix), pattern)
                                    for pattern in patterns):
                                yield (
                                    filename_wo_prefix, filename, open_file,
                                    lambda *args, **kwargs: file_size)
                    else:
                        skipped += 1

                    if (streamed + skipped) % 10000 == 0:
                        logger.info('streaming raw files', streamed=streamed, skipped=skipped)

                    manifest[path(entry)] = {
                        key: entry[key]
                        for key in RawFileQueryResource.manifest_quantities
                        if entry.get(key) is not None
                    }

                if upload_files is not None:
                    upload_files.close()

                logger.info('streaming raw file manifest')
                try:
                    manifest_contents = json.dumps(manifest).encode('utf-8')
                except Exception as e:
                    manifest_contents = json.dumps(
                        dict(error='Could not create the manifest: %s' % (e))).encode('utf-8')
                    logger.error('could not create raw query manifest', exc_info=e)

                yield (
                    'manifest.json', 'manifest',
                    lambda *args: BytesIO(manifest_contents),
                    lambda *args: len(manifest_contents))

            except Exception as e:
                logger.warning(
                    'unexpected error while streaming raw data from query', exc_info=e)

        logger.info('start streaming raw files')
        return streamed_zipfile(
            generator(), zipfile_name='nomad_raw_files.zip', compress=compress)

    @api.doc('post_raw_files_from_query')
    @api.expect(_raw_file_from_query_model)
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.response(200, 'File(s) send', headers={'Content-Type': 'application/zip'})
    @authenticate(signature_token=True)
    def post(self):
        ''' Download a .zip file with all raw-files for all entries that match the given
        search parameters.

        See ``/repo`` endpoint for documentation on the search
        parameters.

        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.

        The zip file will contain a ``manifest.json`` with the repository meta data.
        '''
        patterns: List[str] = None
        try:
            data_in = request.get_json()
            compress = data_in.get('compress', False)
            strip = data_in.get('strip', False)
            pattern = data_in.get('file_pattern', None)
            if isinstance(pattern, str):
                patterns = [pattern]
            elif pattern is None:
                patterns = []
            else:
                patterns = pattern
            query = data_in.get('query', {})
            query_expression = {key: val for key, val in query.items() if '$' in key}
        except Exception:
            abort(400, message='bad parameter types')

        logger = common.logger.bind(query=urllib.parse.urlencode(query, doseq=True))

        search_request = search.SearchRequest()
        apply_search_parameters(search_request, query)
        search_request.include('calc_id', 'upload_id', 'mainfile')

        if query_expression:
            try:
                search_request.query_expression(query_expression)
            except AssertionError as e:
                abort(400, str(e))

        def path(entry):
            return '%s/%s' % (entry['upload_id'], entry['mainfile'])

        calcs = search_request.execute_scan(
            order_by='upload_id',
            size=config.services.download_scan_size,
            scroll=config.services.download_scan_timeout)

        if strip:
            if search_request.execute()['total'] > config.raw_file_strip_cutoff:
                abort(400, 'The requested download has to many files for using "strip".')
            calcs = list(calcs)
            paths = [path(entry) for entry in calcs]
            common_prefix_len = len(utils.common_prefix(paths))
        else:
            common_prefix_len = 0

        def generator():
            try:
                manifest = {}
                directories = set()
                upload_files = None
                streamed, skipped = 0, 0

                for entry in calcs:
                    upload_id = entry['upload_id']
                    mainfile = entry['mainfile']
                    if upload_files is None or upload_files.upload_id != upload_id:
                        logger.info('opening next upload for raw file streaming', upload_id=upload_id)
                        if upload_files is not None:
                            upload_files.close()

                        upload_files = UploadFiles.get(upload_id)

                        if upload_files is None:
                            logger.error('upload files do not exist', upload_id=upload_id)
                            continue

                        def open_file(upload_filename):
                            return upload_files.raw_file(upload_filename, 'rb')

                    upload_files._is_authorized = create_authorization_predicate(
                        upload_id=upload_id, calc_id=entry['calc_id'])
                    directory = os.path.dirname(mainfile)
                    directory_w_upload = os.path.join(upload_files.upload_id, directory)
                    if directory_w_upload not in directories:
                        streamed += 1
                        directories.add(directory_w_upload)
                        for filename, file_size in upload_files.raw_file_list(directory=directory):
                            filename = os.path.join(directory, filename)
                            filename_w_upload = os.path.join(upload_files.upload_id, filename)
                            filename_wo_prefix = filename_w_upload[common_prefix_len:]
                            if len(patterns) == 0 or any(
                                    fnmatch.fnmatchcase(os.path.basename(filename_wo_prefix), pattern)
                                    for pattern in patterns):
                                yield (
                                    filename_wo_prefix, filename, open_file,
                                    lambda *args, **kwargs: file_size)
                    else:
                        skipped += 1

                    if (streamed + skipped) % 10000 == 0:
                        logger.info('streaming raw files', streamed=streamed, skipped=skipped)

                    manifest[path(entry)] = {
                        key: entry[key]
                        for key in RawFileQueryResource.manifest_quantities
                        if entry.get(key) is not None
                    }

                if upload_files is not None:
                    upload_files.close()

                logger.info('streaming raw file manifest')
                try:
                    manifest_contents = json.dumps(manifest).encode('utf-8')
                except Exception as e:
                    manifest_contents = json.dumps(
                        dict(error='Could not create the manifest: %s' % (e))).encode('utf-8')
                    logger.error('could not create raw query manifest', exc_info=e)

                yield (
                    'manifest.json', 'manifest',
                    lambda *args: BytesIO(manifest_contents),
                    lambda *args: len(manifest_contents))

            except Exception as e:
                logger.warning(
                    'unexpected error while streaming raw data from query', exc_info=e)

        logger.info('start streaming raw files')
        return streamed_zipfile(
            generator(), zipfile_name='nomad_raw_files.zip', compress=compress)


def respond_to_get_raw_files(upload_id, files, compress=False, strip=False):
    upload_files = UploadFiles.get(
        upload_id, create_authorization_predicate(upload_id))
    if upload_files is None:
        abort(404, message='The upload with id %s does not exist.' % upload_id)

    if strip:
        common_prefix_len = len(utils.common_prefix(files))
    else:
        common_prefix_len = 0

    try:
        return streamed_zipfile(
            [(
                filename[common_prefix_len:], filename,
                lambda upload_filename: upload_files.raw_file(upload_filename, 'rb'),
                lambda upload_filename: upload_files.raw_file_size(upload_filename)
            ) for filename in files],
            zipfile_name='%s.zip' % upload_id, compress=compress)
    finally:
        upload_files.close()



# Copyright 2017-2018 Berk Onat, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import next
from builtins import object
import numpy as np
import mdtraj
from mdtraj import FormatRegistry
from mdtraj.formats.xtc import load_xtc
from mdtraj.formats.trr import load_trr
from mdtraj.formats.hdf5 import load_hdf5
from mdtraj.formats.lh5 import load_lh5
from mdtraj.formats.netcdf import load_netcdf
from mdtraj.formats.mdcrd import load_mdcrd
from mdtraj.formats.dcd import load_dcd
from mdtraj.formats.binpos import load_binpos
from mdtraj.formats.pdb import load_pdb
from mdtraj.formats.arc import load_arc
from mdtraj.formats.openmmxml import load_xml
from mdtraj.formats.prmtop import load_prmtop
from mdtraj.formats.psf import load_psf
from mdtraj.formats.mol2 import load_mol2
from mdtraj.formats.amberrst import load_restrt, load_ncrestrt
from mdtraj.formats.lammpstrj import load_lammpstrj
from mdtraj.formats.dtr import load_dtr, load_stk
from mdtraj.formats.xyzfile import load_xyz
from mdtraj.formats.hoomdxml import load_hoomdxml
#from mdtraj.formats.tng import load_tng
from mdtraj.core.topology import Topology
from mdtraj.core.trajectory import _TOPOLOGY_EXTS
import mdtraj.formats
import ase.io
import ase.io.formats
from ase.io.formats import format2modulename, import_module
import logging
import os
import sys
logger = logging.getLogger(__name__)

class TrajectoryReader(object):
    """Used to parse various different atomic coordinate files.

    Reading is primarily done by MDTraj and if the file format
    is not supported by MDTraj, ASE is used instead. For file
    formats that are not supported by both codes, a user defined 
    function can be used. See user supported formats in the dictionary.

    Returns all coordinates as numpy arrays.
    """
    def __init__(self):
        self.trajfile = None # File name of the trajectory file with path if needed
        self.topofile = None # File name of the topology file with path if needed
        self.topo_n_atoms = None # Number of atoms at topology file
        self.trajformat = None # Format type of trajectory file
        self.topoformat = None # Format type of topology file
        self.trajhandler = None # The object parsing the trajectory file
        self.trajiter = None # The object parsing the trajectory file
        self.trajchunk = 100 # The object parsing the trajectory file
        self.topohandler = None # The object parsing the topology file
        self.trajcode = None # To explicitly define the parsing library (MDTraj/ASE) for trajectory files
        self.topocode = None # To explicitly define the parsing library (MDTraj/ASE) for topology files
        self.readfirst = False
        self.user_formats = {
            "cif":    {"Crystallographic Information File", self.custom_iread}
        }

    def load(self):
        """Loads the file handles for trajectory and/or topology
        """
        
        if self.topohandler is None:
            if self.topofile:
                self.check_topology_format_support()
        if self.topohandler is None:
            #logger.warning("The topology file format '{}' is not supported by TrajectoryReader.".format(self.topoformat))
            pass

        if self.trajhandler is None:
            self.check_trajectory_format_support()
            if self.trajhandler is None:
                #logger.warning("The trajectory file format '{}' is not supported by TrajectoryReader.".format(self.trajformat))
                #logger.warning("ASE could not read the file '{}' with format '{}'. The contents might be malformed or wrong format used.".format(filename, file_format))
                pass
        return self.trajhandler
    
    def load_topology(self):
        """Loads the file handles for topology only
        """
        
        if self.topohandler is None:
            if self.topofile:
                self.check_topology_format_support()
        if self.topohandler is None:
            logger.error("The topology file format '{}' is not supported by TrajectoryReader.".format(self.topoformat))

        return self.topohandler
    
    def get_file_format(self, filename, givenformat):
        """Returns extension of file
        """
        file_format = None
        if givenformat:
            fname = filename.split("/")[-1]
            # MDTraj expects that extensions start with dot
            if '.' not in fname:
                file_format = "." + fname.lower()
            else:
                file_format = fname.lower()
        else:
            extension = filename.split(".")[-1]
            file_format = "." + extension.lower()
        return file_format

    def check_topology_format_support(self):
        """Check if the given format is supported.
        """
        topofilename = os.path.basename(self.topofile)
        if self.topoformat is None:
            file_format = self.get_file_format(topofilename, self.topoformat)
            self.topoformat = file_format
        else:
            file_format = self.topoformat

        if (file_format in _TOPOLOGY_EXTS and 
            self.topocode is not "ase"):
            self.topohandler = self.load_mdtraj_topology(file_format)
            self.topocode = "mdtraj"

        # If MDTraj does not have support for the format 
        # or can not load the topology, use ASE instead
        ase_support = None
        if self.topohandler is None:
            logger.warning("MDTraj could not read the file '{}' with format '{}'. The contents might be malformed or wrong format used.".format(self.topofile, file_format))
            ase_support = self.get_ase_format_support(file_format)
            # May still have chance that ASE can recognize the 
            # format with its filetype checking function
            if ase_support is None:
                ase_support = ase.io.formats.filetype(topofilename)
                self.topohandler = self.load_ase_support(topofilename, file_format=ase_support)
                self.topocode = "ase"
            else:
                self.topohandler = self.load_ase_support(topofilename, file_format=file_format)
                self.topocode = "ase"

        # May add openbabel and/or Pybel support here.
        if self.topohandler is None:
            logger.error("The format '{}' is not supported by CoordinateReader.".format(file_format))
            return False
        else:
            return True

    def check_trajectory_format_support(self):
        """Check if the given format is supported.
        """
        mdtraj_failed = False
        trajfilename = os.path.basename(self.trajfile)
        if self.trajformat is None:
            file_format = self.get_file_format(trajfilename, self.trajformat)
            self.trajformat = file_format
        else:
            file_format = self.trajformat

        # First check whether MDtraj has support for the file type
        # trajhandler_check = FormatRegistry.loaders[file_format]
        trajhandler_check = None
        try:
            trajhandler_check = FormatRegistry.fileobjects[file_format]
        except KeyError:
            pass
        else:
            if trajhandler_check:
                self.trajhandler = self.mdtraj_iread(mdtraj_handler=trajhandler_check)
                self.trajcode = "mdtraj"

        # If MDTraj does not have support for the format 
        # or can not load the trajectory, use ASE instead
        ase_support = None
        if self.trajhandler is None:
            logger.warning("MDTraj could not read the file '{}' with format '{}'. The contents might be malformed or wrong format used.".format(self.trajfile, file_format))
            ase_support = self.get_ase_format_support(file_format)
            # May still have chance that ASE can recognize the 
            # format with its filetype checking function
            if ase_support is None:
                ase_support = ase.io.formats.filetype(trajfilename)
                self.trajhandler = self.load_ase_support(self.trajfile, file_format=ase_support)
                self.trajcode = "ase"
            else:
                self.trajhandler = self.load_ase_support(self.trajfile, file_format=file_format)
                self.trajcode = "ase"

        # May add openbabel and/or Pybel support here.

        if self.trajhandler is None:
            logger.error("The format '{}' is not supported by CoordinateReader.".format(file_format))
            return False
        else:
            return True

    def iread(self):
        """Returns an iterator that goes through the given trajectory file one
        configuration at a time.
        """
        if self.trajhandler is not None:
            iterator_object = iter(self.trajhandler)
            try:
                while True:
                    self.trajiter = next(iterator_object)
                    return self.trajiter
            except StopIteration:
                pass
            finally:
                del iterator_object
        else:
            return None

    def get_topology(self):
        """Returns an iterator that goes through the given trajectory file one
        configuration at a time.
        """
        return self.topohandler

    def get_unitcell(self):
        """Returns an iterator that goes through the given trajectory file one
        configuration at a time.
        """
        return self.trajhandler.cell_lengths

    def load_mdtraj_topology(self, file_format, **kwargs):
        """Borrowed from mdtraj.trajectory to remove 
           IOError and TypeError warnings
    
        Get the topology from a argument of indeterminate type
        If top is a string, we try loading a pdb, if its a trajectory
        we extract its topology.

        Returns
        -------
        topology : md.Topology
        """

        topology = None
        wrapkwargs = kwargs.copy()
        wrapkwargs.pop("top", None)
        wrapkwargs.pop("atom_indices", None)
        wrapkwargs.pop("frame", None)

        ext=file_format
        top=self.topofile

        if ext in ['.pdb', '.pdb.gz', '.h5','.lh5']:
            _traj = mdtraj.load_frame(top, 0, **wrapkwargs)
            topology = _traj.topology
        elif ext in ['.prmtop', '.parm7']:
            topology = load_prmtop(top, **wrapkwargs)
        elif ext in ['.psf']:
            topology = load_psf(top, **wrapkwargs)
        elif ext in ['.mol2']:
            topology = load_mol2(top, **wrapkwargs).topology
        elif ext in ['.gro']:
            topology = mdtraj.core.trajectory.load_gro(top, **wrapkwargs).topology
        elif ext in ['.arc']:
            topology = load_arc(top, **wrapkwargs).topology
        elif ext in ['.hoomdxml']:
            topology = load_hoomdxml(top, **wrapkwargs).topology
        elif isinstance(top, Trajectory):
            topology = top.topology
        elif isinstance(top, Topology):
            topology = top

        return topology

    def load_ase_support(self, file_name, file_format):
        """Get the topology from a argument of indeterminate type
        If top is a string, we try loading a pdb, if its a trajectory
        we extract its topology.

        Returns
        -------
        topology : generator
        """
        return ase.io.read(file_name, index=":", format=file_format)

    def get_ase_format_support(self, file_format):
        """ Inspired from ase.io.formats.get_ioformat

        Get support level of ASE for file formats

        Returns
        -------
        IOformat list
        """

        ioformat = None
        _format = file_format.replace('-', '_')
        module_name = format2modulename.get(file_format, _format)
        module_handler = None
        module_handler = import_module('ase.io.' + module_name)
        if module_handler:
            module_read = getattr(module_handler, 'read_' + _format, None)
            if module_read and not inspect.isgeneratorfunction(module_read):
                read = functools.partial(wrap_read_function, module_read)
                if module_read:
                    ioformat = all_formats[file_format]

        return ioformat

    def natoms(self):
        """Read the first configuration of the coordinate file to extract the
        number of atoms in it.
        """
        return self.trajiter.xyz.shape[1]

    def ase_iread(self):
        """Wrapper for ASE's iread function. Returns numpy arrays instead of
        Atoms objects.
        """
        # The newest ASE version found in Github has an iread function.
        # After reading the ASE source code, it seems that the ASE iread does
        # actually read the entire file into memory and the yields the
        # configurations from it. Should be checked at some point.
        def ase_generator(iterator):
            """Used to wrap an iterator returned by ase.io.iread so that it returns
            the positions instead of the ase.Atoms object.
            """
            for value in iterator:
                yield value.get_positions()

        iterator = self.trajhandler
        return ase_generator(iterator)

    def mdtraj_iread(self, mdtraj_handler=None, **kwargs):
        """Generator function that is used to read an atomic configuration file (MD
        trajectory, geometry optimization, static snapshot) from a file one frame
        at a time. Only the xyz positions are returned from the file, and no unit
        conversion is done, so you have to be careful with units.

        By using a generator pattern we can avoid loading the entire trajectory
        file into memory. This function will instead load a chunk of the file into
        memory (with MDTraj you can decide the chunk size, with ASE it seems to
        always be one frame), and serve individual files from that chunk. Once the
        frames in one chunk are iterated, the chunk will be garbage collected and
        memory is freed.

        Args:
            filename: String for the file path.
            file_format: String for the file format. If not given the format is
                automatically detected from the extension.

        Yields:
            numpy array containing the atomic positions in one frame.
        """

        # Try to open the file with MDTraj first. With a brief inspection it seems
        # that MDTraj is better performance wise, because it can iteratively load a
        # "chunk" of frames, and still serve the individual frames one by one. ASE
        # on the other hand will iteratively read frames one by one (unnecessary
        # IO).

        # Must use the low level MDTraj API to open files without topology.
        # mdtraj_supported_format = FormatRegistry.loaders[self.fileformat]
        if mdtraj_handler is not None:
            if self.trajchunk == 0:
                try:
                    loader = FormatRegistry.loaders[self.trajformat]
                except KeyError:
                    return
                # If chunk was 0 then we want to avoid filetype-specific code
                # in case of undefined behavior in various file parsers.
                # TODO: this will first apply stride, then skip!
                if self.trajformat not in _TOPOLOGY_EXTS:
                    topkwargs = kwargs.copy()
                    topkwargs['top'] = self.topohandler
                    # standard_names is a valid keyword argument only for files containing topologies
                    topkwargs.pop('standard_names', None)
                    positions = loader(self.trajfile, **topkwargs)
                else:
                    positions = loader(self.trajfile)
                for pos in positions:
                    yield pos
            elif self.trajformat in ('.pdb', '.pdb.gz'):
                # the PDBTrajectortFile class doesn't follow the standard API. Fixing it here
                try:
                    loader = FormatRegistry.loaders[self.trajformat]
                except KeyError:
                    return
                t = loader(filename)
                for i in range(0, len(t), self.trajchunk):
                    positions = t[i:i+self.trajchunk]
                    for pos in positions:
                        yield pos

            else:
                if self.topohandler is None:
                    n_atoms_set=None
                    if self.trajformat in ('.crd', '.mdcrd'):
                        return
                else:
                    n_atoms_set=self.topohandler.n_atoms
                try:
                    with (lambda x: mdtraj_handler(x, n_atoms=n_atoms_set)
                          if self.trajformat in ('.crd', '.mdcrd')
                          else mdtraj_handler(self.trajfile, mode="r"))(self.trajfile) as f:
                        empty = False
                        while not empty:
                            if self.trajformat not in _TOPOLOGY_EXTS:
                                data = f.read_as_traj(self.topohandler, n_frames=self.trajchunk)
                            else:
                                data = f.read_as_traj(n_frames=self.trajchunk)
                            if isinstance(data, tuple):
                                positions = data[0]
                            else:
                                positions = data
                            if len(positions) == 0:
                                empty = True
                            else:
                                for pos in positions:
                                    yield pos
                except IOError:
                    #logger.warning("MDTraj could not read the file '{}' with format '{}'. The contents might be malformed or wrong format used.".format(self.trajfile, self.trajformat))
                    return

    def custom_iread(self, file_handle, format):
        """
        """
        pass

if __name__ == "__main__":
    topo_file = sys.argv[1]
    traj_file = sys.argv[2]
    readdata = TrajectoryReader()
    readdata.trajfile = traj_file
    readdata.topofile = topo_file
#    readdata.trajformat = '.netcdf'
    traj_iterator = readdata.load()
    mytopology = readdata.get_topology()
    # print([ atom.name for atom in mytopology.atoms ])
    readdata.trajchunk=300
    readaccess = False
    # check whether mdcrd can be read with default format
    atom = readdata.iread()
    i=0
    if atom is None:
        # mdcrd trajectory file might be in NetCDF format
        # Trying to read with netcdf format
        readdata = TrajectoryReader()
        readdata.trajfile = traj_file
        readdata.topofile = topo_file
        readdata.trajformat = '.netcdf'
        readdata.trajchunk=300
        traj_iterator = readdata.load()
        mytopology = readdata.get_topology()
        atom = readdata.iread()
    
    if atom is not None:
        readaccess = True
    
#    if readaccess:
#        while True:
#            if atom is not None:
#                print(atom)
#                print(readdata.natoms())
#                i += 1
#            else:
#                break
#            atom = readdata.iread()
    print(i)
#    angles = mdtraj.compute_angles(mytopology)
#    dihedrals = mdtraj.compute_dihedrals(mytopology)
#    print(dihedrals)
    #print(' Chains: %s' % [res for res in mytopology.chains])
    #print(' Residues: %s' % [res for res in mytopology.residues])
    #print(' Atoms: %s' % [res for res in mytopology.atoms])
    #print(' Bonds: %s' % [res for res in mytopology.bonds])
    #print(mytopology.chain(0))
    table, bonds = mytopology.to_dataframe()
    print(bonds)
    print(len(bonds))
    array = table.get("resSeq")
    #print([res for res in array])
    mydict = table.to_dict(orient='list')
    mylist = mydict["resSeq"]
    print('mylist')
    print(mylist)
    print('mylist')
    atomindex = np.arange(len(mylist))
    passarray = np.zeros((len(mylist), 2), dtype=int)
    passarray[:,0] = atomindex
    passarray[:,1] = mylist
    print(table)
    print(passarray)
    #print(bonds)
#    for bond in bonds:
#        print(bond)
    mol_to_mol_bond = []
    mol_to_mol_bond_num = []
    atom_in_mol_bond = []
    atom_in_mol_bond_num = []
    index=0
    print(len([res for res in mytopology.bonds]))
    for res in mytopology.bonds:
        molname1, molatom1 = str(res[0]).split('-')
        molname2, molatom2 = str(res[1]).split('-')
        if molname1 in molname2: 
            atom_in_mol_bond.append(res)
            atom_in_mol_bond_num.append(bonds[index])
        else:
            mol_to_mol_bond.append(res)
            mol_to_mol_bond_num.append(bonds[index])
        index += 1
    print(mol_to_mol_bond)
    print(np.array(mol_to_mol_bond_num))

    atom_list = list(set(mydict["name"]))
    atom_type_dict = {}
    print(atom_list)
    print(len(atom_list))
    masses = {}
    for ielm in range(len(atom_list)):
        elm = atom_list[ielm]
        atom_type_dict.update({elm : ielm})
        for atom in mytopology.atoms:
            if elm == atom.name:
                masses.update({atom.name : atom.element})
    print([atom_list, masses])
    interDict = {}
    interTypeDict = {}
    for elm in atom_list:
        bondList = []
        typeList = []
        bondid = 0
        for bond in mytopology.bonds:
            molname1, molatom1 = str(bond[0]).split('-')
            molname2, molatom2 = str(bond[1]).split('-')
            if (elm == str(molatom1) or elm == str(molatom2)):
                bondList.append(list(bonds[bondid]))
                #typeList.append(list([molatom1,molatom2]))
                typeList.append(list([
                    atom_type_dict[str(molatom1)],
                    atom_type_dict[str(molatom2)]
                    ]))
                interDict.update({elm : bondList})
                interTypeDict.update({elm : typeList})
            bondid += 1

#    print(interDict)
    print(interDict["CH3"])
    print(interTypeDict["CH3"])
    print(len(interDict["CH3"][0]))

#    print([[atom.element.radius, atom.index, atom.name, atom.residue.name, atom.element] for atom in mytopology.atoms])

# Atom labels:
    print(table)
#    for atom in mytopology.atoms:
#        print(atom)
    labellist = mydict["element"]
    #print(labellist)
#    atomlabels = np.arange(len(labellist))
#    labelarray = np.zeros((len(labellist), 2), dtype=int)
#    labelarray[:,0] = atomlabels
#    labelarray[:,1] = labellist
#    print(labelarray)
#    print(readdata.unitcell_lengths)
    latticevectors = readdata.iread().unitcell_vectors
    box = readdata.iread().unitcell_lengths
    step = 0
    for latvec in latticevectors:
        print(step, latvec)
        step += 1
    for pbc in box:
        print(step, box)
        step += 1



            





